import numpy as np
from precise.skaters.covarianceutil.covfunctions import try_invert
from precise.skaters.portfolioutil.portfunctions import portfolio_variance
from typing import List
from itertools import zip_longest
from precise.skaters.locationutil.vectorfunctions import normalize


# Long-short min-var portfolios where the only constraint is sum(w)=1


def unit_portfolio_factory(pre=None, cov=None):
    """ Signed min var portfolio summing to unity """
    if pre is not None:
        return unitary_from_pre(pre=pre)
    else:
        return unitary_from_cov(cov=cov)


def unitary_from_pre(pre):
    """ Signed min var portfolio summing to unity """
    # No optimization required
    n_dim = np.shape(pre)[1]
    wones = np.ones(shape=(n_dim, 1))
    w = normalize(np.squeeze(np.matmul(pre, wones)))
    return np.array(w)


def unitary_from_cov(cov):
    pre = try_invert(cov)
    return unitary_from_pre(pre=pre)


def unitary_portfolio_variance(cov=None, pre=None):
    """
        Variance of the unit min-var portfolio
        (Used in some hierarchical methods to allocate capital)
    """
    w = unit_portfolio_factory(pre=pre, cov=cov)
    return portfolio_variance(cov=cov,w=w)


def prc_unit_alloc(covs:List, pres:List=None)->[float]:
    """ Allocate capital between portfolios using either cov or pre matrices
    :param covs:  List of covariance matrices
    :param pres:  List of precision matrices
    :return: Capital allocation vector
    """
    if pres is None:
        pres = []
    return normalize([ 1/unitary_portfolio_variance(cov=cov, pre=pre) for cov, pre in zip_longest(covs, pres, fillvalue=None) ])






import numpy as np
import math
import time


def historical_log_likelihood(pre, xs, lb, mu=None):
    """ Log likelihood of matrix of data """
    from sklearn.covariance._empirical_covariance import log_likelihood, empirical_covariance
    if mu is None:
        emp_cov = empirical_covariance(xs)
    else:
        emp_cov = empirical_covariance(xs - mu)
    ll = log_likelihood(emp_cov=emp_cov, precision=pre)
    if ll<lb:
        ll=lb # Do something?!
    return ll


def vector_log_likelihood(pre, y, lb):
    """ Log likelihood of y
    :param cov:    Inverse of pre, the predicted precision
    :param pre:
    :param y:      realized data points
    :param lb:     lower bound, returned if cov is degen
    :return:
    """
    # c.f. sklearn
    #     term2 = 1/2 fast_logdet(pre)
    #     log_likelihood_  = -np.sum(emp_cov * precision) + fast_logdet(precision)
    #     log_likelihood_ -= p * np.log(2 * np.pi)
    # Derivation at
    # https://stats.stackexchange.com/questions/351549/maximum-likelihood-estimators-multivariate-gaussian
    from sklearn.utils.extmath import fast_logdet
    p = len(y)
    logdet = fast_logdet(pre)
    if logdet < lb:
        return lb
    else:
        ll_term_1 = - (p / 2) * math.log(2 * math.pi)
        ll_term_2 =  (1 / 2) * logdet
        y_innovation = np.atleast_2d(y)
        y_innovation_t = y_innovation.transpose()
        ll_term_3 = - 1 / 2 * np.matmul(np.matmul(y_innovation, pre), y_innovation_t)[0, 0]
        ll = ll_term_1 + ll_term_2 + ll_term_3
    return ll


def cov_likelihood(contestant, xs, n_burn=10, lb=-1000, ub=1000):
    # Used as evaluator
    return cov_skater_loglikelihood(f=contestant, xs=xs, with_metrics=True, n_burn=n_burn, lb=lb, ub=ub)


def cov_skater_loglikelihood(f, xs, n_burn=10, with_metrics=True, lb=-1000, ub=1000, verbose=True):
    """
        Gaussian likelihood of a cov skater applied to data xs

    :param f:  cov skater
    :param lb, ub  lower and upper bounds for ll of individual point
    :param xs:
    :return:
    """
    start_time = time.time()
    inv_time = 0

    n_obs, n_dim = np.shape(xs)
    assert n_obs>n_burn
    s = {}

    if verbose:
        print('   burn in for '+str(n_burn))
    for y in xs[:n_burn]:
        y_hat, y_cov, s = f(s=s, y=y, k=1, e=-1)

    if verbose:
        print('   evaluating  '+str(n_obs-n_burn))

    ll = 0
    y_hat_prev = None
    y_cov_prev = None
    for m,y in enumerate( xs[n_burn:]):
        if y_hat_prev is not None:
            # Evaluate last prediction
            inv_start_time = time.time()
            try:
                y_inv_prev = np.linalg.inv(y_cov_prev)
            except np.linalg.LinAlgError:
                y_inv_prev = np.linalg.pinv(y_cov_prev)
            inv_time += time.time()-inv_start_time

            dy = np.array(y) - np.array(y_hat_prev)
            ll_delta = vector_log_likelihood(pre=y_inv_prev, y=dy, lb=lb)
            if ll_delta>ub:
                ll_delta = ub
            ll += ll_delta

        # Store predictions for assessment against next data point
        y_hat_prev = y_hat
        y_cov_prev = y_cov

        # Make next prediction
        y_hat, y_cov, s = f(s=s, y=y, k=1, e=1)

    total_time = time.time()-start_time
    metrics = {'total time':total_time,'inversion time':inv_time,'time':total_time-inv_time}
    return ll, metrics if with_metrics else ll


def pre_skater_loglikelihood(f, xs, n_burn=10, with_metrics=True, lb=-1000, ub=1000):
    """
        Gaussian likelihood of a precision skater applied to data xs

    :param f:
    :param lb, ub  lower and upper bounds for ll of individual point
    :param xs:
    :return:
    """
    start_time = time.time()

    n_obs, n_dim = np.shape(xs)
    assert n_obs>n_burn
    s = {}

    for y in xs[:n_burn]:
        y_hat, y_pre, s = f(s=s,y=y,k=1)

    ll = 0
    y_hat_prev = None
    y_pre_prev = None
    for m,y in enumerate( xs[n_burn:]):
        if y_hat_prev is not None:
            dy = np.array(y) - np.array(y_hat_prev)
            ll_delta = min( vector_log_likelihood(pre=y_pre_prev, y=dy, lb=lb),ub )
            ll += ll_delta

        # Store predictions for assessment against next data point
        y_hat_prev = y_hat
        y_pre_prev = y_pre

        # Make next prediction
        y_hat, y_pre, s = f(s=s, y=y, k=1)


    total_time = time.time()-start_time
    metrics = {'total time':total_time}
    return ll, metrics if with_metrics else ll



if __name__=='__main__':
    from precise.skaters.covariance.runemp import run_emp_pcov_d0
    xs = np.random.randn(500,3)
    ll, metrics = cov_skater_loglikelihood(f=run_emp_pcov_d0, xs=xs, with_metrics=True)
    print(ll)
    from pprint import pprint
    pprint(metrics)









from precise.skaters.portfoliostatic.diagportfactory import diagonal_portfolio_factory


def diag_long_port(cov=None, pre=None):
    return diagonal_portfolio_factory(cov=cov, pre=pre)


DIAG_LONG_PORT = [ diag_long_port ]
DIAG_PORT = DIAG_LONG_PORT


# Ack: https://arxiv.org/pdf/2108.12627.pdf

import numpy as np
import math
from scipy.optimize import newton


def huber_mean(xs:[[float]], a:float=1.0, b=2.0, n_iter=20, atol=1e-8)->[float]:
    """ Compute a columnwise pseudo-mean of xs, by minimizing a generalized Huber error that is
        proportional to x^2 near zero and asymptotes to |x| as |x|->infinity.
               f(x) = 1/a log( exp(a*(x-mu)) + exp(-(a*(x-mu)) + b )
        This is the same as the function below, except the parameter a will multiply std(x)
        :param xs:    (n_samples, n_vars)
        :param a:    Generalized Huber parameter as per formula
        :param b:    Generalized Huber parameter as per formula above, scalar or (nvars,)
        :param with_fraction_converged: If True, will return the mean number of rows where convergence was achieved
        :return:      (n_vars,)   location parameters

    """
    if len(xs)==0:
        raise ValueError('Cannot compute huber mean for no data')
    elif len(xs)==1:
        return xs[0]
    else:
        x_std = np.nanstd(xs,axis=0)
        xs_rescaled = xs/(10*x_std)
        a_abs = a/10.
        scaled_mean = huber_mean_absolute_params(xs=xs_rescaled, a=a_abs, b=b, n_iter=n_iter, atol=atol, with_fraction_converged=False)
        unscaled_mean = scaled_mean * x_std*10
        return unscaled_mean


def huber_mean_absolute_params(xs:[[float]], a, b, n_iter=20, atol=1e-8, with_fraction_converged=False)->[float]:
    """ Finds the generalized Huber locations for many variables at once
        Each column of xs represents a different variable whose pseudo-mean will be computed
        Thus the result mu might be compared to np.mean(xs, axis=0)

        The function being minimized w.r.t mu is
              f(x) = 1/a log( exp(a*(x-mu)) + exp(-(a*(x-mu)) + b )
        and for |x|->0 this has asymptote:
             f(x) ->  log(2+b)/a + a/(2+b) * (x-mu)^2
        whereas for |x|->infinity
             f(x) -> |x-mu|

        This Huber function is not the standard Huber loss https://en.wikipedia.org/wiki/Huber_loss
        Rather, it is based on https://arxiv.org/pdf/2108.12627.pdf

    :param xs:    (n_samples, n_vars)
    :param a:    Generalized Huber parameter as per formula above, scalar or (nvars,)
    :param b:    Generalized Huber parameter as per formula above, scalar or (nvars,)
    :return:      (n_vars,)   location parameters
    """
    x_median = np.median(xs, axis=0)
    x_mean = np.mean(xs, axis=0)
    lb = np.where(x_median < x_mean, x_median, x_mean)-0.01
    ub = np.where(x_median > x_mean, x_median, x_mean)+0.01
    x0 = 0.75*lb+0.25*ub
    flb = huber_deriv(lb,a,b,xs)
    fub = huber_deriv(ub,a,b,xs)
    mu = newton(func=huber_deriv, x0=x0, args=(a, b, xs), tol=1e-6, rtol=1e-4 )
    return mu


def huber_deriv(mu, a, b, xs):
    """ Derivative of generalized Huber loss w.r.t. mu

         f'(x) = 1/a log( exp(a*(x-mu)) + exp(-(a*(x-mu)) + b )

    :param mu:   (n_samples,)                    # Vector of location parameters
    :param xs :   (n_samples,n_vars)              # Data
    :param a:    Generalized Huber parameter(s)
    :param b:    Generalized Huber parameter(s)
    :return:
    """
    n_samples, n_vars = np.shape(xs)
    mu_rep = np.tile(np.atleast_2d(mu), (n_samples,1))
    y = xs - mu_rep
    chain_rule = -1.0
    numer = np.exp(np.dot(y,a) ) - np.exp(-np.dot(a,y))  # (n_samples, n_params)
    denom = numer + b
    gradient = chain_rule*numer/denom
    return np.mean(gradient,axis=0)


def huber_abs_error(mu, a, b, xs):
    """ Generalized Huber loss which is "like" abs error, as it approaches |x-mu| as |x-mu|-> infinity
          f(x) = 1/a log( exp(a*(x-mu)) + exp(-(a*(x-mu)) + b )
     """
    n_samples, n_vars = np.shape(xs)
    mu_rep = np.tile(np.atleast_2d(mu), (n_samples, 1))
    y = xs - mu_rep
    numer = np.log(np.exp(np.dot(y, a)) + np.exp(np.dot(a, y)) + b)  # (n_samples, n_params)
    denom = a
    return numer / denom


def mean_huber_linear_error(mu, a, b, xs):
    """
        This convenience function returns columnwise mean.
    """
    f_val = huber_abs_error(mu=mu, a=a, b=b, xs=xs)
    return np.mean(f_val, axis=0)


def huber_squared_error(mu, a, b, xs):
    """ Rescaled generalized Huber loss which is "like" squared error, in
    the sense that it approaches (x-mu)^2 as |x-mu|-> 0

        If
         f(x) = 1/a log( exp(a*(x-mu)) + exp(-(a*(x-mu)) + b )
        Then
         f(x) -> log(2+b)/a + a/(2+b) * x^2   as x->0
        by Taylor. So we define
         g(x) :=  ( f(x) - log(2+b)/a ) * (2+b)/a

    """
    f = huber_abs_error(mu=mu,a=a,b=b,xs=xs)
    c = math.log(2 + b) / a
    d = a / (2 + b)
    g = ( f - c ) / d
    return g


def mean_huber_squared_error(mu, a, b, xs):
    """
        This convenience function returns columnwise mean.
    """
    f_val = huber_squared_error(mu=mu, a=a, b=b, xs=xs)
    return np.mean(f_val, axis=0)


def mean_quadratic_error(mu, xs):
    """
       Here as a useful comparison to the below
    """
    n_samples, n_vars = np.shape(xs)
    mu_rep = np.tile(np.atleast_2d(mu), (n_samples, 1))
    y = xs - mu_rep
    return np.nanmean(y**2, axis=0)





# TODO: Move into ratings package to remove this dependency

def elo_expected(d :float ,f :float =400 )->float:
    """ Expected points scored in a match by White player
    :param d:   Difference in rating (Black minus White)
    :param f:   "F"-Factor
    :return:
    """
    if d/ f > 8:
        return 0.0
    elif d / f < -8:
        return 1.0
    else:
        return 1. / (1 + 10 ** (d / f))


def elo_change(white_elo, black_elo, points, k=25, f=400):
    """
    :param white_elo:
    :param black_elo:
    :param points:     1 if White wins, 0 if Black wins
    :param k:   K-Factor, how much to update
    :param f:   F-Factor
    :return:
    """
    d = black_elo - white_elo
    expected_points = elo_expected(d=d, f=f)
    w = points - expected_points
    return k*w, -k*w


import numpy as np
from precise.skaters.covarianceutil.covfunctions import try_invert, approx_diag_of_inv
from precise.skaters.locationutil.vectorfunctions import normalize
from precise.skaters.portfoliostatic.equalport import equal_long_port

# Use only the diagonal elements of cov


def diagonal_portfolio_factory(pre=None, cov=None, sparse_approx=False, ridge=0.1):
    """ Min-var portfolio using only diagonal entries in cov
        (e.g. precision weighting of stacked models)
    :param pre:  precision matrix
    :param cov:  covariance matrix
    :param sparse_approx: Use fast approximate inversion of precision matrix
    :return:
    """
    if pre is not None:
        return diagonal_from_pre(pre=pre, sparse_approx=sparse_approx, ridge=ridge)
    else:
        return diagonal_from_cov(cov=cov, ridge=ridge)


def diagonal_from_pre(pre, sparse_approx=False, ridge=0.1):
    """ Signed min var portfolio summing to unity """
    if sparse_approx:
        cov = np.diag( approx_diag_of_inv(pre) )
    else:
        cov = try_invert(pre)
    return diagonal_from_cov(cov=cov, ridge=ridge)


def diagonal_from_cov(cov, ridge=0.1):
    d = np.diag(cov)
    if any([di<1e-6 for di in d]):
        return equal_long_port(cov=cov)
    else:
        mean_cov = np.mean(d)
        return normalize( [ 1/(d+ridge*mean_cov) for d in np.diag(cov)] )






from precise.skaters.managers.rflmanagerfactory import rfl_hrp_factory


# Risk Measures available in RiskFolio-Lib
# (See this example https://nbviewer.org/github/dcajasn/Riskfolio-Lib/blob/master/examples/Tutorial%2024.ipynb)
#
# 'vol': Standard Deviation.
# 'MV': Variance.
# 'MAD': Mean Absolute Deviation.
# 'MSV': Semi Standard Deviation.
# 'FLPM': First Lower Partial Moment (Omega Ratio).
# 'SLPM': Second Lower Partial Moment (Sortino Ratio).
# 'VaR': Conditional Value at Risk.
# 'CVaR': Conditional Value at Risk.
# 'EVaR': Entropic Value at Risk.
# 'WR': Worst Realization (Minimax)
# 'MDD': Maximum Drawdown of uncompounded cumulative returns (Calmar Ratio).
# 'ADD': Average Drawdown of uncompounded cumulative returns.
# 'DaR': Drawdown at Risk of uncompounded cumulative returns.
# 'CDaR': Conditional Drawdown at Risk of uncompounded cumulative returns.
# 'EDaR': Entropic Drawdown at Risk of uncompounded cumulative returns.
# 'UCI': Ulcer Index of uncompounded cumulative returns.
# 'MDD_Rel': Maximum Drawdown of compounded cumulative returns (Calmar Ratio).
# 'ADD_Rel': Average Drawdown of compounded cumulative returns.
# 'DaR_Rel': Drawdown at Risk of compounded cumulative returns.
# 'CDaR_Rel': Conditional Drawdown at Risk of compounded cumulative returns.
# 'EDaR_Rel': Entropic Drawdown at Risk of compounded cumulative returns.
# 'UCI_Rel': Ulcer Index of compounded cumulative returns.


def rfl_hrp_var_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='MV', n_buffer=200)


def rfl_hrp_vol_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='vol', n_buffer=200)


def rfl_hrp_mad_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='MAD', n_buffer=200)


def rfl_hrp_msv_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='MSV', n_buffer=200)


def rfl_hrp_flpm_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='FLPM', n_buffer=200)


def rfl_hrp_slpm_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='SLPM', n_buffer=200)


def rfl_hrp_var_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='VaR', n_buffer=200)


def rfl_hrp_cvar_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='CVaR', n_buffer=200)


def rfl_hrp_evar_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='EVaR', n_buffer=200)


def rfl_hrp_wr_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='WR', n_buffer=200)


def rfl_hrp_mdd_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='MDD', n_buffer=200)


def rfl_hrp_add_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='ADD', n_buffer=200)


def rfl_hrp_dar_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='DaR', n_buffer=200)


def rfl_hrp_cdar_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='CDaR', n_buffer=200)


def rfl_hrp_edar_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='EDaR', n_buffer=200)


def rfl_hrp_mddrel_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='MDD_Rel', n_buffer=200)


def rfl_hrp_addrel_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='ADD_Rel', n_buffer=200)


def rfl_hrp_cdarrel_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='CDaR_Rel', n_buffer=200)


def rfl_hrp_edarrel_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='EDaR_Rel', n_buffer=200)


def rfl_hrp_ucirel_long_manager_n200(y, s, k=1, e=1, **ignore):
    return rfl_hrp_factory(y=y, s=s, k=k, e=e, metric='UCI_Rel', n_buffer=200)


RFL_HRP_LONG_MANAGERS = [rfl_hrp_var_long_manager_n200, rfl_hrp_vol_long_manager_n200,
                         rfl_hrp_mad_long_manager_n200, rfl_hrp_msv_long_manager_n200,
                         rfl_hrp_flpm_long_manager_n200, rfl_hrp_slpm_long_manager_n200,
                         rfl_hrp_var_long_manager_n200, rfl_hrp_cvar_long_manager_n200,
                         rfl_hrp_evar_long_manager_n200, rfl_hrp_wr_long_manager_n200,
                         rfl_hrp_mdd_long_manager_n200, rfl_hrp_add_long_manager_n200,
                         rfl_hrp_dar_long_manager_n200, rfl_hrp_cdar_long_manager_n200,
                         rfl_hrp_edar_long_manager_n200, rfl_hrp_mddrel_long_manager_n200,
                         rfl_hrp_addrel_long_manager_n200, rfl_hrp_cdarrel_long_manager_n200,
                         rfl_hrp_edarrel_long_manager_n200, rfl_hrp_ucirel_long_manager_n200]

RFL_HRP_LS_MANAGERS = []
RFL_HRP_MANAGERS = RFL_HRP_LONG_MANAGERS + RFL_HRP_LS_MANAGERS

RFL_LONG_MANAGERS = RFL_HRP_LONG_MANAGERS
RFL_LS_MANAGERS = RFL_HRP_LS_MANAGERS






import numpy as np
import math
import pandas as pd
from precise.skaters.covarianceutil.pdutil import square_to_square_dataframe, square_to_column_series, square_to_index_series

# Functions acting on cov, corrcoef matrices and other square matrices
# If square pd.DataFrame are supplied instead, index and columns are preserved


def cov_to_corrcoef(a):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, cov_to_corrcoef)
    else:
        variances = np.diagonal(a)
        denominator = np.sqrt(variances[np.newaxis, :] * variances[:, np.newaxis])
        return a / denominator


def normalize(x):
    if isinstance(x,dict):
        return normalize_dict_values(x)
    else:
        try:
            return x/sum(x)
        except TypeError:
            return [xi/sum(x) for xi in x]


def normalize_dict_values(d):
    return dict( zip(d.keys(), normalize(list(d.values())) ))


def multiply_diag(a, phi, copy=True):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(df=a, cov_func=multiply_diag, phi=phi, copy=copy)
    else:
        if copy:
            b = np.copy(a)
            return multiply_diag(b, phi=phi, copy=False)
        else:
            n = np.shape(a)[0]
            for i in range(n):
                a[i, i] = a[i, i] * phi
            return a


def grand_mean(a):
    # np.trace(a)/n might be faster, haven't checked
    if isinstance(a,pd.DataFrame):
        return grand_mean(a.values)
    else:
        return np.mean(a.diagonal())


def grand_shrink(a, lmbd, copy=True):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, grand_shrink, a=a, lmbd=lmbd)
    else:
        if copy:
            B = np.copy(a)
            return grand_shrink(B, lmbd=lmbd, copy=False)
        else:
            n = np.shape(a)[0]
            mu = grand_mean(a)
            return (1-lmbd) * a + lmbd * mu * np.eye(n)


def affine_inversion(a, phi=1.01, lmbd=0.01, copy=True):
    """ Combination of "ridge" (multiply diagonal) and "shrinkage" (towards mu*I)
    :param a:
    :param phi:
    :param lmbd:
    :return:
    """
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, affine_inversion, phi=phi, lmbd=lmbd)
    else:
        shrunk_cov = affine_shrink(a=a, phi=phi, lmbd=lmbd)
        try:
            pre = np.linalg.inv(shrunk_cov)
        except np.linalg.LinAlgError:
            pre = np.linalg.pinv(shrunk_cov)
        return pre


def affine_shrink(a, phi=1.01, lmbd=0.01, copy=True):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, affine_shrink, phi=phi, lmbd=lmbd)
    else:
        ridge_cov = multiply_diag(a, phi=phi, copy=copy)
        shrunk_cov = grand_shrink(ridge_cov, lmbd=lmbd, copy=copy)
        return shrunk_cov


def is_symmetric(a, rtol=1e-05, atol=1e-08):
    if isinstance(a, pd.DataFrame):
        return is_symmetric(a.values, rtol=rtol, atol=atol)
    else:
        return np.allclose(a, a.T, rtol=rtol, atol=atol)


def to_symmetric(a):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, to_symmetric)
    else:
        return (a + a.T) / 2.0


def dense_weights_from_dict(d:dict, shape=None, n_dim:int=None):
    """ Convert {1:3.141,0:123} -> [ 123, 3.141 ]
    :param d: Dictionary whose keys should be 0..n
    :param shape:
    :param n_dim:
    :return:
    """
    if shape is None:
        if n_dim is not None:
            shape = (n_dim,)
        else:
            n_dim = max(d.values())+1
    w = np.ndarray(shape=shape)
    for i in range(n_dim):
        w[i] = d[i]
    return w


def nearest_pos_def(a):
    """Find the nearest positive-definite matrix to input

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which
    credits [2].

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, nearest_pos_def)
    else:
        b = to_symmetric(a)
        _, s, V = np.linalg.svd(b)
        H = np.dot(V.T, np.dot(np.diag(s), V))
        a2 = (b + H) / 2
        a3 = (a2 + a2.T) / 2

        if is_positive_def(a3):
            return a3

        spacing = np.spacing(np.linalg.norm(a))
        # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
        # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
        # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
        # for `np.spacing`), we use the above definition. CAVEAT: our `spacing`
        # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
        # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
        # `spacing` will, for Gaussian random matrixes of small dimension, be on
        # othe order of 1e-16. In practice, both ways converge, as the unit test
        # below suggests.
        I = np.eye(a.shape[0])
        k = 1
        while not is_positive_def(a3):
            mineig = np.min(np.real(np.linalg.eigvals(a3)))
            a3 += I * (-mineig * k**2 + spacing)
            k += 1

        return a3


def is_positive_def(a):
    """Returns true when input is positive-definite, via Cholesky"""
    if isinstance(a,pd.DataFrame):
        return is_positive_def(a.values)
    else:
        try:
            _ = np.linalg.cholesky(a)
            return True
        except np.linalg.LinAlgError:
            return False


def make_diagonal(a):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, make_diagonal)
    else:
        return np.diag(np.diag(a))


def mean_off_diag(a):
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, mean_off_diag)
    else:
        n = np.shape(a)[0]
        b = np.vectorize(int)(a)
        b = b - np.eye(n)
        the_sum = np.sum(b,axis=None)
        return the_sum/(n*(n-1))


def corr_distance(corr, expon=0.5):
    """
        Convert correlations into a distance between variables
    """
    if isinstance(corr, pd.DataFrame):
        return square_to_square_dataframe(corr, corr_distance, expon=expon)
    else:
        return ((1 - np.array(corr)) / 2.) ** expon


def cov_distance(cov, expon=0.5):
    """
           Convert covariance into a distance between variables
    """
    if isinstance(cov, pd.DataFrame):
        return square_to_square_dataframe(cov, cov_distance, expon=expon)
    else:
        corr = cov_to_corrcoef(cov)
        return corr_distance(corr=corr, expon=expon)


def try_invert(a, **affine_inversion_kwargs):
    """
       Attempt to invert a matrix by whatever means, falling back to ridge + shrinkage as required
    """
    if isinstance(a, pd.DataFrame):
        return square_to_square_dataframe(a, try_invert, **affine_inversion_kwargs)
    else:
        try:
            return np.linalg.inv(a)
        except np.linalg.LinAlgError:
            try:
                return np.linalg.pinv(a)
            except np.linalg.LinAlgError:
                return affine_inversion(a, **affine_inversion_kwargs)


def weaken_cov(cov, diag_multipliers:[float], off_diag_additional_factor=0.9):
    """  Augment a covariance matrix
    :param cov:
    :param diag_multipliers:             Vector to multiply diagonals by
    :param off_diag_additional_factor:   Additional multiplicative factor
    :return:
    """
    if isinstance(cov, pd.DataFrame):
        return square_to_square_dataframe(cov, weaken_cov, diag_multipliers=diag_multipliers, off_diag_additional_factor=off_diag_additional_factor)
    else:
        covs = np.copy(cov)
        for i, di in enumerate(diag_multipliers):
            covs[i, i] = covs[i, i] * di
            for j, dj in enumerate(diag_multipliers):
                if j != i:
                    covs[i, j] = off_diag_additional_factor * math.sqrt(di * dj) * covs[i, j]
        return covs


def approx_diag_of_inv(a):
    """
        Approximate diagonal entries of the inverse of a matrix
    """
    # TODO: https://cholmod-extra.readthedocs.io/en/latest/functions.html#sparse-inverse
    raise NotImplementedError


def bottom_schur_complement(A, B, C, D, gamma=1.0):
    # D - gamma C A^{-1} B
    return schur_complement(A=D, B=C, C=B, D=A, gamma=gamma)


def schur_complement(A,B,C,D, gamma=1.0):
    # A - gamma B D^{-1} C
    return _schur_complement_solve(A=A, B=B, C=C, D=D, gamma=gamma)


def _schur_complement_solve(A, B, C, D, gamma, warn=False, throw=False):
    # A - B D^{-1} C
    DinvC = inverse_multiply(a=D, b=C, warn=warn, throw=throw)
    M = A - gamma*np.dot( B, DinvC )

    CAREFUL = warn or throw
    if CAREFUL:
        rankD = np.linalg.matrix_rank(D)
        dimD = max(np.shape(D))
        if rankD==dimD:
            checkM = _schur_complement_pseudo(A=A,B=B,C=C,D=D, gamma=gamma)
            if not np.allclose( M, checkM ):
                print('schurly not')
        elif warn:
            print(' D is rank deficient, so schur complement is method dependent ')

    return M


def _schur_complement_pseudo(A, B, C, D, gamma):
    return A - gamma*np.dot( np.dot(B, np.linalg.pinv(D)),C)


def _schur_complement_direct(A, B, C, D, gamma):
    return A - gamma*np.dot( np.dot(B, np.linalg.pinv(D)),C)


def inverse_multiply(a, b, warn=False, throw=False):
    # Want  x = a^{-1} b
    #       a x = b
    #       x = solve(a,b)
    x = np.linalg.solve(a, b)
    if (warn or throw):
        if np.linalg.matrix_rank(x)<max(np.shape(a)):
            print('a is rank deficient so result is method dependent ')
        else:
            x_ = np.dot(np.linalg.inv(a), b)
            if not np.allclose(x_, x):
                if warn:
                    print('schurly not inverse multiply')
                if throw:
                    raise ValueError
    return x


def multiply_by_inverse(a, b, throw=True):
    #  Want x = a b^{-1}
    #       xt = bt^{-1} at  = inverse_multiply(bt, at)
    #       bt xt = at
    #       xt = solve(bt, at)
    if not np.allclose( np.shape(b)[1], np.shape(b)[0], np.shape(a)[1]):
        raise ValueError('dims wrong')
    bt = np.transpose(b)
    at = np.transpose(a)
    xt = np.linalg.solve(bt,at)
    x  = np.transpose(xt)
    x_check = np.dot(a, np.linalg.inv(b))
    if not np.allclose(x, x_check ) and throw:
        raise Exception('schurly not')
    return x





import numpy as np
import random


def create_correlated_dataset(n, mu=None, dependency=None, scale=None):
    if mu is None:
        mu = (2.2, 4.4, 1.5)
    if dependency is None:
        dependency = np.array([[0.2, 0.5, 0.7],[0.3, 0.2, 0.2],[0.5,0.3,0.1]])
    if scale is None:
        scale = (1, 5, 3)
    latent = np.random.randn(n, dependency.shape[0])
    dependent = latent.dot(dependency)
    scaled = dependent * scale
    x = scaled + mu
    return x







from precise.skaters.covarianceutil.conventions import X_TYPE
from precise.skaters.location.averagingpre import averager


def emp(s:dict=None, x:X_TYPE=None):
    """ Empirical mean """
    return averager(s=s, x=x, method='emp')


from precise.whereami import BATTLE_RESULTS_DIR
from glob import glob
import json
from collections import Counter
import os

# Retrieving collections of battles


def win_dirs(genre='likelihood'):
    return [ d[:-1] for d in glob(BATTLE_RESULTS_DIR + os.path.sep + genre + "/*/", recursive = False)]


def win_data(genre='likelihood'):
    """
    :return: [ ( category, counter ) ]
    """
    return [(category, load_win_data(cat_files)) for (category,cat_files) in win_files(genre=genre)]


def win_files(genre='likelihood'):
    return [(wd.split('/')[-1], glob(wd+"/*.json")) for wd in win_dirs(genre=genre)]


def load_win_data(cat_files):
    data = Counter()
    for fn in cat_files:
        with open(fn,'rt') as fh:
            new_data = Counter(json.load(fh))
            data.update(new_data)
    return data



if __name__=='__main__':
    wd = win_data(genre='likelihood')
    print([c for c,_ in win_data()])




from runthis import parse_kwargs
from precise.skatervaluation.battleutil.arrangingbattles import cov_likelihood_battle
import os

# Infers params from file name and runs a battle

if __name__=='__main__':
    params = parse_kwargs(__file__.split(os.path.sep)[-1])
    cov_likelihood_battle(params=params)


from precise.skaters.covariance.bufhuberfactory import buf_huber_d0_factory
from precise.skaters.covarianceutil.differencing import d1_factory


def buf_huber_pcov_d0_a1_b2_n50(y, s, k=1,e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=1.0, b=2.0, n_buffer=50, e=e)


def buf_huber_pcov_d0_a05_b2_n50(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=2.0, n_buffer=50, e=e)


def buf_huber_pcov_d0_a1_b5_n50(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=5.0, n_buffer=50, e=e)


def buf_huber_pcov_d0_a1_b2_n100(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=1.0, b=2.0, n_buffer=100, e=e)


def buf_huber_pcov_d0_a05_b2_n100(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=2.0, n_buffer=100, e=e)


def buf_huber_pcov_d0_a1_b5_n100(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=5.0, n_buffer=100, e=e)


def buf_huber_pcov_d0_a1_b2_n200(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=1.0, b=2.0, n_buffer=200, e=e)


def buf_huber_pcov_d0_a05_b2_n200(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=2.0, n_buffer=200, e=e)


def buf_huber_pcov_d0_a1_b5_n200(y, s, k=1, e=1):
    assert k==1
    return buf_huber_d0_factory(y=y,s=s,a=0.5, b=5.0, n_buffer=200, e=e)


BUF_HUBER_D0_COV_SKATERS = [buf_huber_pcov_d0_a1_b2_n50, buf_huber_pcov_d0_a05_b2_n50, buf_huber_pcov_d0_a1_b5_n50,
                            buf_huber_pcov_d0_a1_b2_n100, buf_huber_pcov_d0_a05_b2_n100, buf_huber_pcov_d0_a1_b5_n100,
                            buf_huber_pcov_d0_a1_b2_n200, buf_huber_pcov_d0_a05_b2_n200, buf_huber_pcov_d0_a1_b5_n200]


def buf_huber_pcov_d1_a1_b2_n50(y, s, k=1, e=1):
    return d1_factory(y=y,s=s,k=k,a=1.0, b=2.0, n_buffer=50, e=e)


def buf_huber_pcov_d1_a1_b2_n100(y, s, k=1, e=1):
    return d1_factory(y=y,s=s,k=k,a=1.0, b=2.0, n_buffer=100, e=e)


BUF_HUBER_D1_COV_SKATERS = [buf_huber_pcov_d1_a1_b2_n50, buf_huber_pcov_d1_a1_b2_n100]



from precise.skaters.portfolioutil.portfunctions import portfolio_variance
import numpy as np
import pandas as pd
from collections import Counter
from pprint import pprint


def points_race( ranker, ranker_kwargs, n_iter=100, n_top=20):
    c = Counter()
    for _ in range(n_iter):
        rankings = ranker(as_frame=False, **ranker_kwargs)
        points = [ (port_name, max(len(rankings)-k,0) ) for k,(port_name,_,_) in enumerate(rankings) ]
        c.update(dict(points))
        pprint(c.most_common(n=n_top))
    return c


def stock_portfolio_variance_points_race(n_iter=100,n_top=50, **kwargs):
    return points_race(n_iter=n_iter, n_top=n_top, ranker=stock_portfolio_variance_rankings, ranker_kwargs=kwargs)


def equity_portfolio_variance_points_race(n_iter=100,n_top=50, **kwargs):
    return points_race(n_iter=n_iter, n_top=n_top, ranker=m6_equity_portfolio_variance_rankings, ranker_kwargs=kwargs)


def equity_portfolio_correlation_points_race(n_iter=100,n_top=50, **kwargs):
    return points_race(n_iter=n_iter, n_top=n_top, ranker=m6_equity_portfolio_correlation_rankings, ranker_kwargs=kwargs)


def stock_portfolio_variance_rankings(ports, n_dim=10, n_obs = 300, k=1, as_frame=True, n_iter=10):
    """
        Quick and dirty
    """
    from precise.skatertools.data.equityhistorical import random_cached_equity_dense
    data = random_cached_equity_dense(n_dim=n_dim, n_obs=n_obs, verbose = False, k=k)
    t_obs = int(0.75*n_obs)
    test_cov = np.cov(data[:t_obs] ,rowvar=False)
    train_cov = np.cov(data[t_obs:], rowvar=False)
    return portfolio_variance_rankings(cov_test=test_cov, ports=ports, cov_train=train_cov, as_frame=as_frame, n_iter=n_iter )


def m6_equity_portfolio_variance_rankings(ports, n_dim=10, n_obs = 300, interval='d', etf=1, as_frame=True, n_iter=10):
    """
        Quick and dirty
    """
    from precise.skatertools.data.equitylive import random_m6_returns
    data = random_m6_returns(n_dim=n_dim, n_obs=n_obs, verbose = False, interval = interval, etf = etf)
    t_obs = int(0.5*n_obs)
    test_cov = np.cov(data[:t_obs] ,rowvar=False)
    train_cov = np.cov(data[t_obs:], rowvar=False)
    return portfolio_variance_rankings(cov_test=test_cov, ports=ports, cov_train=train_cov, as_frame=as_frame, n_iter=n_iter )


def m6_equity_portfolio_correlation_rankings(ports, n_dim=10, n_obs = 300, interval='d', etf=1, as_frame=True):
    """
        Quick and dirty comparison using empirical corrcoef for cov
    """
    from precise.skatertools.data.equitylive import random_m6_returns
    data = random_m6_returns(n_dim=n_dim, n_obs=n_obs, verbose = False, interval = interval, etf = etf)
    t_obs = int(0.666*n_obs)
    test_cov = np.corrcoef(data[:t_obs] ,rowvar=False)
    train_cov_gen = np.corrcoef(data[t_obs:], rowvar=False)
    n = np.shape(data)[1]
    ys = np.random.multivariate_normal(mean=np.zeros(n), cov=train_cov_gen, size=50 )
    train_cov = np.cov(ys, rowvar=False)

    return portfolio_variance_rankings(cov_test=test_cov, ports=ports, cov_train=train_cov, as_frame=as_frame )


def portfolio_variance_rankings(cov_train, ports, cov_test=None, as_frame=True, n_iter=1):
    """  Really crude comparison with single train/test

    :param cov_train:    Training cov matrix
    :param ports:        List of portfolio methods
    :param cov_test:
    :param as_frame:
    :return:
    """

    if cov_test is None:
        cov_test = cov_train

    rankings = list()
    for port in ports:
        w = port(cov_train)
        try:
            in_var = portfolio_variance(cov=cov_train, w=w)
            out_var = portfolio_variance(cov=cov_test, w=w)
        except ValueError:
            print('portfolio may be wrong shape, or cov_tain, cov_test')
            raise
        rankings.append((port.__name__, in_var, out_var))
    leaderboard = sorted(rankings, key=lambda t :t[2])

    if as_frame:
        bestout = leaderboard[0][2]
        df = pd.DataFrame(columns=['method' ,'invar' ,'outvar'] ,data=leaderboard)
        df['out sample relative'] = df['outvar']/bestout
        return df
    else:
        return leaderboard





from precise.skaters.managers.managerfactory import static_cov_manager_factory_d0
from precise.skaters.covariance.ewapm import ewa_pm_factory
from precise.skaters.portfoliostatic.schurportfactory import schur_portfolio_factory
from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
from precise.skaters.portfoliostatic.weakalloc import weak_allocation_factory
from precise.skaters.covariance.ewaempfactory import ewa_emp_pcov_factory
from functools import partial


def schur_weak_pm_manager_factory(y, s, target, n_emp, e, r, a=1.0, b=None, n_split=5, gamma=0.0, delta =0.0, a_alloc=1.0, b_alloc=None):
    """
       HRP weak portfolio construction using partial moments cov estimation
           a, b             - weak coefs used at leaf
           a_alloc, b_alloc - weak coefs used at parents
    """
    f = partial( ewa_pm_factory, k=1, r=r,target=target, n_emp=n_emp )
    alloc = partial( weak_allocation_factory, a=a_alloc, b=b_alloc )
    leaf_port = partial( weak_portfolio_factory, a=a, b=b )
    hrp_port = partial(schur_portfolio_factory, alloc=alloc, port=leaf_port, n_split=n_split, gamma=gamma, delta=delta)
    return static_cov_manager_factory_d0(f=f, port=hrp_port, y=y, s=s, e=e)


def schur_weak_ewa_manager_factory(y, s, n_emp, e, r, a=1.0, b=None, n_split=5, gamma=0.0, delta=0.0, a_alloc=1.0, b_alloc=None):
    """
       HRP weak portfolio construction using partial moments cov estimation
           a, b             - weak coefs used at leaf
           a_alloc, b_alloc - weak coefs used at parents
    """
    f = partial( ewa_emp_pcov_factory, k=1,r=r, n_emp=n_emp )
    alloc = partial( weak_allocation_factory, a=a_alloc, b=b_alloc )
    leaf_port = partial( weak_portfolio_factory, a=a, b=b )
    hrp_port = partial(schur_portfolio_factory, alloc=alloc, port=leaf_port, n_split=n_split, gamma=gamma, delta=delta)
    return static_cov_manager_factory_d0(f=f, port=hrp_port, s=s, y=y, e=e)





from precise.skaters.location.empricalpre import emp
import numpy as np


def emp_d0(y, s:dict, k=1, **ignore):
    """ Rudimentary empirical skater with no cov estimation """
    # See precision.covariance.empirical.run_emp_pcov_d1 if you want cov estimates
    s = emp(s=s, x=np.array(y))
    y_hat = s['mean']
    return y_hat, np.eye(len(y)), s


def emp_d1(y, s:dict, k=1, **ignore):
    """ Rudimentary differenced empirical skater with no cov estimation """
    # See precision.covariance.empirical.run_emp_pcov_d1 if you want cov estimates
    if not s:
        s = {'prev_y':y,'dy':{}}
        return np.zeros_like(y), np.eye(len(y)), s
    else:
        dy = y - s['prev_y']
        dy_hat, _, s['dy'] = emp_d0(y=dy, s=s['dy'], k=1)
        y['prev_y'] = y
        y_hat = s['prev_y'] + dy_hat + s['dy']['v']['mean']
        y_cov = np.eye(len(y))
        return y_hat, y_cov, s





from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory


# Fast long-only approximately min-var portfolios where the only constraints are sum(w)=1, w>0

def weak_long_port(cov=None, pre=None):
    """
          Approximately min-var portfolio using optimized weakening parameter b
    """
    return weak_portfolio_factory(cov=cov, pre=pre, a=1.0, b=None, with_neg_mass=False)



WEAK_LONG_PORT = [weak_long_port]
WEAK_LS_PORT = []
WEAK_PORT = WEAK_LONG_PORT + WEAK_LS_PORT




import time
import numpy as np
from momentum.functions import var_init, var_update
from itertools import zip_longest


def manager_info(contestant, xs, n_burn, **ignore):
    # evaluator for battles
    metrics = manager_stats(mgr=contestant, xs=xs, n_burn=n_burn)
    return metrics['info'], metrics


def manager_var(contestant, xs, n_burn, **ignore):
    # evaluator for battles
    metrics = manager_stats(mgr=contestant, xs=xs, n_burn=n_burn)
    return -metrics['var'], metrics


def manager_stats(mgr, xs, n_burn=10):
    """
       Compute manager stats.
       Sends e=-1 during burn-in, then e=1 during assessment period

       :returns  dict with  mean, var, kurtosis etc of portfolio returns
    """
    start_time = time.time()
    n_obs, n_dim = np.shape(xs)
    assert n_obs > n_burn

    s = {}     # Manager state

    print('     burning in')
    for y in xs[:n_burn]:
        w, s = mgr(s=s, y=y, k=1, e=-1)

    print('     evaluating')
    metrics = var_init()
    w_prev = None
    mrg_time = 0
    for m, y in enumerate(xs[n_burn:]):
        if w_prev is not None:
            x = sum( [ yi*wi for yi,wi in zip_longest( y,w_prev) ] )
            metrics = var_update(metrics, x=x)

        # Store portfolio for assessment against next data point
        w_prev = w

        # Make next portfolio selection
        st = time.time()
        w, s = mgr(s=s, y=y, k=1, e=1)
        mrg_time += time.time()-st

    total_time = time.time() - start_time
    metrics.update({'time':mrg_time,'total_time':total_time})
    metrics['info'] = metrics['mean']/metrics['std']
    print('     info = '+str(metrics['info']))
    print('     std  = ' + str(metrics['std']))
    return metrics


if __name__=='__main__':
    import random
    from precise.skaters.managers.allmanagers import LONG_MANAGERS
    xs = np.random.randn(500,3)
    for contestant in LONG_MANAGERS:
       print(contestant.__name__)
       ll, metrics = manager_info(contestant=contestant, xs=xs, n_burn=50)
       print('   '+str(ll))



from precise.skaters.portfoliostatic.diagallocfactory import diagonal_allocation_factory


def diag_alloc(covs=None, pres=None):
    return diagonal_allocation_factory(covs=covs, pres=pres)


DIAG_ALLOC = [ diag_alloc ]


import numpy as np
import pandas as pd
from precise.skaters.covarianceutil.covfunctions import try_invert
from precise.skaters.covarianceutil.pdutil import dict_or_series_to_scalar, square_and_vector_to_scalar, vector_to_vector
from precise.skaters.locationutil.vectorfunctions import normalize

# Some common operations and functionals of portfolios
# Here the portfolio w can be a list, ndarray, dict or Series


def positive_mass(w):
    if isinstance(w,(dict,pd.Series)):
        return dict_or_series_to_scalar(d=w, func=positive_mass )
    else:
        return sum([wi for wi in w if wi>0])


def negative_mass(w):
    if isinstance(w, (dict, pd.Series)):
        return dict_or_series_to_scalar(d=w, func=negative_mass)
    else:
        return sum([-wi for wi in w if wi<0])


def relative_negative_mass(w):
    if isinstance(w, (dict, pd.Series)):
        return dict_or_series_to_scalar(d=w, func=relative_negative_mass)
    else:
        pm = positive_mass(w)
        nm = negative_mass(w)
        return nm/(pm-nm)


def portfolio_variance(w, cov=None, pre=None):
    if cov is None:
        cov = try_invert(a=pre)
    if isinstance(cov,pd.DataFrame):
        return square_and_vector_to_scalar(a=cov, w=w, func=portfolio_variance)
    else:
        w1 = np.atleast_2d(w)
        wt = np.atleast_2d(w).transpose()
        return np.matmul( np.matmul(w1, np.array(cov)), wt)[0][0]


def normalize_portfolio(w):
    """ Ensure sum(w) = 1 """
    if isinstance(w,(dict,pd.Series)):
        return vector_to_vector(w=w, func=normalize)
    else:
        return normalize(w)


def exclude_negative_weights(w, with_neg_mass=False):
    """
         Take out negative weights, preserving presumed total mass
         :param w   1d array, list, Series or dict
    """
    if isinstance(w,(dict,pd.Series)):
        assert with_neg_mass==False
        return vector_to_vector(w=w, func=exclude_negative_weights, with_neg_mass=False)
    else:
        pos_mass = sum([wi for wi in w if wi>0])
        neg_mass = sum([-wi for wi in w if wi<0])
        if neg_mass>1e-10:
            presumed_mass = pos_mass-neg_mass
            ratio = presumed_mass/pos_mass
            w_pos = np.array([ wi*ratio if wi>0 else 0.0 for wi in w])
        else:
            w_pos = np.array( [ wi/pos_mass for wi in w] )
        return (w_pos, neg_mass) if with_neg_mass else w_pos


def var_scaled_returns(cov, mu:float, r:float):
    import warnings
    warnings.filterwarnings('error')
    vars = np.diag(cov)
    typical_var = np.mean(vars)
    try:
        result = r+(mu-r)*np.array([ v/typical_var for v in vars])
    except RuntimeWarning as e:
        print(e)
        pass

    return result


if __name__=='__main__':
    sgma = np.array([[1,4],[4,20]])
    w = [0.3,0.4]
    print( portfolio_variance(cov=sgma,w=w) )


from precise.skaters.covariance.ewaempfactory import ema_scov
from precise.skaters.covarianceutil.covfunctions import grand_mean, grand_shrink
from sklearn.covariance._shrunk_covariance import ledoit_wolf_shrinkage
import numpy as np


# EWALD short for "Exp Weight Avg" Ledoit-Wolf 
# ---------------------------------------------
# Experimental online estimator inspired by Ledoit-Wolf
# Keeps a buffer of last n_buffer observations
# Tracks quantities somewhat akin to a^2, d^2 used by LW to
# estimate a "reasonable" linear shrinkage


def ewa_lw_scov_factory(s, y, r, e=1):
    s = lw_ema_scov(s=s,x=y,r=r)
    x = np.copy(s['ema_scov']['mean'])
    x_cov = np.copy(s['scov'])
    return x, x_cov, s


def lw_ema_scov(s:dict, x=None, r=0.025)->dict:
    if s.get('ema_scov') is None:
        if isinstance(x,int):
            return _lw_ema_scov_init(n_dim=x, r=r)
        else:
            s = _lw_ema_scov_init(n_dim=len(x), r=r)
    if x is not None:
        s = _lw_ema_scov_update(s=s, x=x, r=r)
    return s


def _lw_ema_scov_init(n_dim, r):
    sc = ema_scov({}, n_dim, r=r)
    return {'ema_scov':sc,
            'bn_bar':None,
            'a2':0,
            'mn':0,
            'n_new':0,
            'buffer':[]}


def _lw_ema_scov_update(s, x, r):
    """
        Attempts to track quantities similar to those used to estimate LD shrinkage
    """
    # Uses buffered LD up to 2*n_emp observations, then switches to an updating scheme
    x = np.asarray(x)
    s['ema_scov']  = ema_scov(s=s['ema_scov'], x=x, r=r)
    s['buffer'].append(x)
    if len(s['buffer'])>s['ema_scov']['n_emp']:
        # Update running estimate of the LD shrinkage parameter
        s['n_new'] = s['n_new']+1
        xl = s['buffer'].pop(0)
        xc = np.atleast_2d(xl-s['ema_scov']['mean'])  # <--- Do we need this?
        scov = s['ema_scov']['scov']

        # Compute d^2
        mn = grand_mean(scov)
        s['mn'] = mn
        n_dim = np.shape(scov)[0]
        s['dn'] = np.linalg.norm((scov - mn * np.eye(n_dim))**2)

        # Update b^2
        xc2 = xc**2
        xl2 = np.dot(xc2.T,xc2) - scov
        if s.get('bn_bar') is None:
            s['bn_bar'] = s['lmbd']*s['dn']
            s['lmbd_lw'] = 1.0 * s['lmbd']
        r_shrink = r/2 # <--- Heuristic
        bk = np.linalg.norm( xl2 )
        s['bn_bar'] = (1-r_shrink)*s['bn_bar'] + r_shrink*bk    # b^2
        ratio = bk/s['dn']

        # Imply new shrinkage
        bn = min( s['bn_bar'], s['dn'] )
        lmbd = bn/s['dn']
        s['lmbd'] = lmbd

    # If still warming up, override lmbd with traditional Ledoit-Shrinkage
    if 2 < s['ema_scov']['n_samples'] < 2*s['ema_scov']['n_emp']:
        X = np.asarray(s['buffer'])
        s['lmbd'] = ledoit_wolf_shrinkage(X=X)

    # Create shrunk version of moving avg sample covariance
    if s['ema_scov']['n_samples']>2:
        scov = s['ema_scov']['scov']
        s['scov'] = grand_shrink(a=scov, lmbd=s['lmbd'], copy=True)
    else:
        s['scov'] = np.eye(len(x))

    return s




















from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory


def unit_port(cov=None, pre=None):
    return unit_portfolio_factory(cov=cov, pre=pre)


UNIT_LS_PORT = [unit_port]
UNIT_LONG_PORT = []
UNIT_PORT = UNIT_LONG_PORT + UNIT_LS_PORT


from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
from precise.skaters.portfoliostatic.acdcfactory import schur_complement_portfolio_factory

# Acca Dacca Portfolio method
#
# It's gonna rock you ... eventually


def acdc_weak_s5_port(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=weak_portfolio_factory, cov=cov, pre=pre, n_split=5)


def acdc_weak_s25_port(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=weak_portfolio_factory, cov=cov, pre=pre, n_split=25)


ACDC_PORT = []   # Not working yet
ACDC_LONG_PORT = []



from precise.skaters.covarianceutil.conventions import X_TYPE
from precise.skaters.covariance.bufskfactory import buf_sk_factory
from precise.skaters.covarianceutil.differencing import d1_factory
from sklearn.covariance import GraphicalLasso, GraphicalLassoCV, LedoitWolf, MinCovDet, OAS, EmpiricalCovariance


# Sklearn

def buf_sk_emp_pcov_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Sklearn empirical covariance based estimator for IID observations """
    assert k == 1
    return buf_sk_factory(cls=EmpiricalCovariance, y=y, s=s, n_buffer=100, n_emp=2, e=e)


def buf_sk_emp_pcov_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Sklearn empirical covariance based estimator for IID observations """
    assert k == 1
    return buf_sk_factory(cls=EmpiricalCovariance, y=y, s=s, n_buffer=200, n_emp=2, e=e)



# Ledoit-Wolf

def buf_sk_ld_pcov_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Ledoit-Wolf covariance based estimator for IID observations """
    assert k == 1
    return buf_sk_factory(cls=LedoitWolf, y=y, s=s, n_buffer=100, n_emp=5, e=e)


def buf_sk_ld_pcov_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Ledoit-Wolf covariance based estimator for IID observations """
    assert k == 1
    return buf_sk_factory(cls=LedoitWolf, y=y, s=s, n_buffer=200, n_emp=5, e=e)


def buf_sk_ld_pcov_d0_n300(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Ledoit-Wolf covariance based estimator for IID observations """
    assert k == 1
    return buf_sk_factory(cls=LedoitWolf, y=y, s=s, n_buffer=300, n_emp=5, e=e)


# Graphical Lasso variants

def buf_sk_gl_pcov_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200}
    return buf_sk_factory(cls=GraphicalLasso, y=y, s=s, n_buffer=100, n_emp=50, cls_kwargs=cls_kwargs)


def buf_sk_gl_pcov_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200}
    return buf_sk_factory(cls=GraphicalLasso, y=y, s=s, n_buffer=200, n_emp=50, cls_kwargs=cls_kwargs, e=e)


# Graphical Lasso with Cross Validation


def buf_sk_glcv_pcov_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=100, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_d0_n100_t0(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200, 'assume_centered':True}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=100, n_emp=30, cls_kwargs=cls_kwargs, e=e)



def buf_sk_glcv_pcov_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=200, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_d0_n200_t0(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200, 'assume_centered':True}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=200, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_d0_n300(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=300, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_lars_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV LARS model covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200, 'mode': 'lars'}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=100, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_lars_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV LARS model covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200, 'mode': 'lars'}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=200, n_emp=30, cls_kwargs=cls_kwargs, e=e)


def buf_sk_glcv_pcov_lars_d0_n300(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ Graphical Lasso CV LARS model covariance based estimator for IID observations """
    assert k == 1
    cls_kwargs = {'tol': 1e-4, 'max_iter': 200, 'mode': 'lars'}
    return buf_sk_factory(cls=GraphicalLassoCV, y=y, s=s, n_buffer=300, n_emp=30, cls_kwargs=cls_kwargs, e=e)


# Minimum Cov Det


def buf_sk_mcd_pcov_d0_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ Minimumn covariance based estimator for IID observations """
    assert k == 1
    # https://scikit-learn.org/stable/modules/generated/sklearn.covariance.MinCovDet.html#sklearn.covariance.MinCovDet
    return buf_sk_factory(cls=MinCovDet, y=y, s=s, n_buffer=n_buffer, n_emp=10, e=e)


# Oracle approximation

def buf_sk_oas_pcov_d0_n100(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ OAS based estimator for IID observations """
    # https://arxiv.org/abs/0907.4698
    assert k == 1
    return buf_sk_factory(cls=OAS, y=y, s=s, n_buffer=100, n_emp=10, e=e)


def buf_sk_oas_pcov_d0_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ OAS based estimator for IID observations """
    # https://arxiv.org/abs/0907.4698
    assert k == 1
    return buf_sk_factory(cls=OAS, y=y, s=s, n_buffer=200, n_emp=10, e=e)


def buf_sk_oas_pcov_d0_n300(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ OAS based estimator for IID observations """
    # https://arxiv.org/abs/0907.4698
    assert k == 1
    return buf_sk_factory(cls=OAS, y=y, s=s, n_buffer=300, n_emp=10, e=e)


# -----------------------------------------------------------------------------
# Differenced versions to be applied to time-series with independent increments
# -----------------------------------------------------------------------------

def buf_sk_ec_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    assert k == 1
    return d1_factory(f=buf_sk_emp_pcov_d0_n100, y=y, s=s, n_buffer=n_buffer, e=e)


def buf_sk_ld_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ Ledoit-Wolf covariance based estimator for IID changes """
    assert k == 1
    return d1_factory(f=buf_sk_ld_pcov_d0_n100, y=y, s=s, n_buffer=n_buffer, k=1, e=e)


def buf_sk_oas_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ OAS based estimator for IID changes  """
    assert k == 1
    return d1_factory(f=buf_sk_oas_pcov_d0_n100, y=y, s=s, n_buffer=100, k=1, e=e)


def buf_sk_oas_pcov_d1_n200(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ OAS based estimator for IID changes  """
    assert k == 1
    return d1_factory(f=buf_sk_oas_pcov_d0_n100, y=y, s=s, n_buffer=200, k=1, e=e)


def buf_sk_oas_pcov_d1_n300(y: X_TYPE = None, s: dict = None, k=1, e=1):
    """ OAS based estimator for IID changes  """
    assert k == 1
    return d1_factory(f=buf_sk_oas_pcov_d0_n100, y=y, s=s, n_buffer=300, k=1, e=e)


def buf_sk_gl_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ GL based estimator for IID changes  """
    assert k == 1
    return d1_factory(f=buf_sk_gl_pcov_d0_n100, y=y, s=s, n_buffer=n_buffer, e=e)


def buf_sk_glcv_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ GL CV based estimator for IID changes  """
    assert k == 1
    return d1_factory(f=buf_sk_glcv_pcov_d0_n100, y=y, s=s, n_buffer=n_buffer, k=1, e=e)


def buf_sk_mcd_pcov_d1_n100(y: X_TYPE = None, s: dict = None, n_buffer: int = 100, k=1, e=1):
    """ Minimumn covariance based estimator for IID changes """
    assert k == 1
    # https://scikit-learn.org/stable/modules/generated/sklearn.covariance.MinCovDet.html#sklearn.covariance.MinCovDet
    return d1_factory(f=buf_sk_mcd_pcov_d0_n100, y=y, s=s, n_buffer=n_buffer, k=1, e=e)


BUF_SK_D0_SKATERS = [buf_sk_emp_pcov_d0_n100, buf_sk_emp_pcov_d0_n100,
                     buf_sk_ld_pcov_d0_n100, buf_sk_ld_pcov_d0_n200, buf_sk_ld_pcov_d0_n300,
                     buf_sk_gl_pcov_d0_n100, buf_sk_gl_pcov_d0_n200,
                     buf_sk_glcv_pcov_d0_n100, buf_sk_glcv_pcov_d0_n200, buf_sk_glcv_pcov_d0_n300,
                     buf_sk_glcv_pcov_d0_n100_t0, buf_sk_glcv_pcov_d0_n200_t0, buf_sk_glcv_pcov_d0_n300,
                     buf_sk_glcv_pcov_lars_d0_n100, buf_sk_glcv_pcov_lars_d0_n200, buf_sk_glcv_pcov_lars_d0_n300,
                     buf_sk_mcd_pcov_d0_n100,
                     buf_sk_oas_pcov_d0_n100, buf_sk_oas_pcov_d0_n200, buf_sk_oas_pcov_d0_n300]

BUF_SK_D1_SKATERS = [buf_sk_ec_pcov_d1_n100, buf_sk_ld_pcov_d1_n100, buf_sk_gl_pcov_d1_n100, buf_sk_glcv_pcov_d1_n100,
                     buf_sk_mcd_pcov_d1_n100,
                     buf_sk_oas_pcov_d1_n100]

if __name__ == '__main__':
    from precise.whereami import url_from_skater_name
    print(url_from_skater_name(name='buf_sk_mcd_pcov_d1_n100'))



from precise.skaters.covariance.runempfactory import emp_pcov
import numpy as np

# Running empirical covariance with infinite memory, equal weighting

def run_emp_pcov_d0(y, s:dict, k=1, a=None, t=None, e=1):
    """
        Empirical covariance skater that assumes y's are iid

    :param y:  (n_dim,)   Incoming vector of observations
    :param s:             State
    :param k:             Steps ahead to predict (this won't change anything)
    :param a:             Variables know in advance (ignored)
    :param t:             Epoch time (ignored)
    :param e:             Allocated computation time (ignored)
    :return:  (n_dim,), (n_dim,n_dim), dict
              Point estimate, cov estimate, posterior state
    """
    assert k==1
    s = emp_pcov(s=s,x=y)
    x = s['mean']
    x_cov = s['pcov']
    return x, x_cov, s


def run_emp_pcov_d1(y, s:dict, k=1, a=None, t=None, e=None):
    """
       Empirical covariance skater that assumes changes in y's are iid
    """
    assert k==1
    if not s or s.get('dy'):
        s = {'prev_y':y,
             'dy':{}}
        return y, np.eye(len(y)), s
    else:
        dy = y-s['prev_y']
        dy_hat, dy_cov, s['dy'] = run_emp_pcov_d0(y=dy, s=s['dy'], k=1)
        y['prev_y'] = y
        x = s['prev_y'] + dy_hat
        return x, dy_cov, s


RUN_EMP_DO_COV_SKATERS = [run_emp_pcov_d0]
RUN_EMP_D1_COV_SKATERS = [run_emp_pcov_d1]





from heapq import heappush ,heappop, heapify ,_heapify_max
from typing import Union


# Running scalar median
# Ack: https://medium.com/mind-boggling-algorithms/streaming-algorithms-running-median-of-an-array-using-two-heaps-cd1b61b3c034


def med(s,x:Union[float,int]=None)->dict:
    """ Running median
       :param x: scalar float
    """
    if not s or s.get('low_heap') is None:
        s = dict()
        s['low_heap'] = []
        s['high_heap'] = []
        s['median'] = 0
    if x < s['median']:
        heappush(s['low_heap'], x)
        _heapify_max(s['low_heap'])
    else:
        heappush(s['high_heap'], x)

    if len(s['low_heap']) > len(s['high_heap'] ) +1:
        heappush(s['high_heap'], heappop(s['low_heap']))
        _heapify_max(s['low_heap'])
    elif len(s['high_heap']) > len(s['low_heap']) + 1:
        heappush(s['low_heap'], heappop(s['high_heap']))
        _heapify_max(s['low_heap'])

    if len(s['low_heap']) == len(s['high_heap']):
        s['median'] = float(s['low_heap'][0] + s['high_heap'][0] ) /2.0
    else:
        s['median'] = float(s['low_heap'][0]) if len(s['low_heap']) > len(s['high_heap']) else float(s['high_heap'][0])
    return s


if __name__=='__main__':
    import numpy as np
    import random
    xs = np.random.randn(random.choice([1,5,10,1000]))
    s = {}
    for x in xs:
        s = med(s=s,x=x)
    assert s['median'] == np.median(xs)



from precise.skaters.covariance.ewalwfactory import ewa_lw_scov_factory
from precise.skaters.covarianceutil.differencing import d1_factory

# EWALD short for "Exp Weight Avg" Ledoit-Wolf

# Experimental estimator inspired by Ledoit-Wolf
# Keeps a buffer of last n_buffer observations
# Tracks quantities akin to a^2, d^2 used by LW to
# estimate a "reasonable" linear shrinkage


def ewa_lw_scov_d0_r01(y, s, k=1, e=1):
    assert k==1
    return ewa_lw_scov_factory(y=y, s=s, r=0.01, e=e)


def ewa_lw_scov_d0_r02(y, s, k=1, e=1):
    assert k==1
    return ewa_lw_scov_factory(y=y, s=s, r=0.02)


EWA_LW_D0_COV_SKATERS = [ewa_lw_scov_d0_r01, ewa_lw_scov_d0_r02 ]


def ewa_lw_scov_d1_r01(y, s, k=1, e=1):
    assert k == 1
    return d1_factory( f=ewa_lw_scov_factory, y=y, s=s, r=0.01)


def ewa_lw_scov_d1_r02(y, s, k=1, e=1):
    assert k == 1
    return d1_factory( f=ewa_lw_scov_factory, y=y, s=s, r=0.02)


EWA_LW_D1_COV_SKATERS = [ewa_lw_scov_d1_r01, ewa_lw_scov_d1_r02 ]


import numpy as np
import pandas as pd
from precise.skaters.covarianceutil.covfunctions import to_symmetric, dense_weights_from_dict, normalize_dict_values, nearest_pos_def
from precise.skaters.covarianceutil.covfunctions import try_invert
from precise.skaters.locationutil.vectorfunctions import normalize
from precise.skaters.portfolioutil.portfunctions import portfolio_variance
from pypfopt import EfficientFrontier
from typing import List
from itertools import zip_longest
from precise.skaters.portfolioutil.portfunctions import var_scaled_returns
from pypfopt.exceptions import OptimizationError
from cvxpy.error import SolverError
try:
    from scipy.sparse.linalg import ArpackNoConvergence
except ImportError:
    from scipy.sparse.linalg.eigen import ArpackNoConvergence

from precise.skaters.covarianceutil.covfunctions import affine_shrink

# Thin wrapper for PyPortfolioOpt
# For full flexibility refer to the package https://pyportfolioopt.readthedocs.io/en/latest/MeanVariance.html


PPO_METHODS = ['max_sharpe','min_volatility','max_quadratic_utility']
PPO_LONG_BOUNDS = (0, 1)
PPO_UNIT_BOUNDS = (-1, 1)


def ppo_sharpe_port(cov=None, pre=None, as_dense=True):
    """ Max Sharpe ratio portfolio using cov-implied returns
    :param cov:        Covariance matrix
    :param pre:        Precision matrix
    :param as_dense:   If false, will return weights in dict formet
    :return: np.array of weights
    """
    return ppo_portfolio_factory(method='max_sharpe', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_vol_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='min_volatility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_quad_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_quadratic_utility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_sharpe_ls_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_sharpe', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


def ppo_vol_ls_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='min_volatility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


def ppo_quad_ls_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_quadratic_utility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


def ppo_portfolio_factory(method:str, cov=None, pre=None, as_dense=False, weight_bounds=None,
                          risk_free_rate:float=0.02, mu:float=0.04, n_attempts=5, warn=False, throw=False):
    """
    :param method:
    :param cov:
    :param pre:
    :param as_dense:         If set to True, will force return of np.array even if supplied dataframe
    :param weight_bounds:
    :return:  Can return a dictionary of variable names and weights
    """

    expected_returns = var_scaled_returns(cov=cov,mu=mu,r=risk_free_rate)

    if weight_bounds is None:
        weight_bounds = PPO_LONG_BOUNDS

    if cov is None:
        cov = try_invert(pre)

    # Set return style
    as_series = (not as_dense) and isinstance(cov,pd.DataFrame)

    # Tidy up cov and send to optimizer ... repeatedly with more shrinkage as needed
    shrunk_cov = nearest_pos_def( to_symmetric( np.copy(cov) ) )
    converged = False
    warned = False
    for attempt_no in range(n_attempts):
        n_dim = np.shape(cov)[0]
        ef = EfficientFrontier(expected_returns=expected_returns, cov_matrix=shrunk_cov,
                               weight_bounds=weight_bounds)
        port_method = getattr(ef, method)
        try:
            if method=='max_sharpe':
                port_method(risk_free_rate=risk_free_rate)
            else:
                port_method()
            converged = True
        except (OptimizationError, SolverError, ArpackNoConvergence, UserWarning):
            converged = False
        if converged:
            break
        else:
            warned = True
            if warn:
                print('    warning: '+method+' did not converge on attempt '+str(attempt_no))
            shrunk_cov = affine_shrink(a=shrunk_cov,phi=1.02, lmbd=0.01, copy=False)

    if not converged:
        if throw:
            raise NotImplementedError('pyportfolio opt failed even after shrinkage')
        else:
            print('   PyPortfolioOpt failed ... falling back to minimum variance')
            from precise.skaters.portfoliostatic.equalport import equal_long_port
            return equal_long_port(cov=cov, as_dense=not as_series)

    if converged and warned:
        if warn:
            print('       ... but with shrinkage it converges okay.')

    weights = ef.clean_weights()
    weights = normalize_dict_values(weights)

    if as_series:
        return pd.Series( index=list(weights.keys()), data=list(weights.values()) )
    else:
        return dense_weights_from_dict(weights, n_dim=n_dim)


def long_from_cov( cov, as_dense=True ):
    """ Backward compat """
    return ppo_vol_port(cov=cov, as_dense=as_dense)


def long_from_pre(pre, as_dense=True):
    """ Backward compate """
    return ppo_vol_port(pre=pre, as_dense=as_dense)


def ppo_portfolio_variance(method:str, cov=None, pre=None):
    """
        Variance of the unit min-var portfolio
        (Used in some hierarchical methods to allocate capital)
    """
    w = ppo_portfolio_factory(method=method, pre=pre, cov=cov)
    return portfolio_variance(cov=cov,w=w)


def ppo_sharpe_alloc(covs:List, pres:List)->[float]:
    return _ppo_portfolio_allocation(method='max_sharpe', covs=covs, pres=pres)


def ppo_vol_alloc(covs:List, pres:List)->[float]:
    return _ppo_portfolio_allocation(method='min_volatility', covs=covs, pres=pres)


def ppo_quad_alloc(covs:List, pres:List)->[float]:
    return _ppo_portfolio_allocation(method='max_quadratic_utility', covs=covs, pres=pres)


def _ppo_portfolio_allocation(method:str, covs:List, pres:List)->[float]:
    """ Allocate capital between portfolios using either cov or pre matrices
    :param covs:  List of covariance matrices
    :param pres:  List of precision matrices
    :return: Capital allocation vector
    """
    return normalize([ 1/ppo_portfolio_variance(method=method, cov=cov, pre=pre) for cov, pre in zip_longest(covs, pres, fillvalue=None) ])





def both_cov(s):
    """ Ensure tracking object has both population and sample cov """
    if s.get('n_samples')>1:
        if s.get('scov') is None and s['pcov'] is not None:
            s['scov'] = s['n_samples']/(s['n_samples']-1)*s['pcov']
        if s.get('pcov') is None and s['scov'] is not None:
            s['pcov'] = (s['n_samples'] - 1)/ s['n_samples'] * s['scov']
    return s



import numpy as np
from precise.skaters.portfoliostatic.unitportfactory import unitary_from_cov
from precise.skaters.covarianceutil.covfunctions import multiply_by_inverse

# Provides the transforms necessary to reduce
#
#       min  w^t C R w  subject to w^t c = 1
#
# to the special case R=I, c=1

# (So yeah, nothing really to do with unitary operators, just transforms of the unitary portfolio problem)


def scaled_unitary_problem(C, c):
    """ Find equivalent problem to:

          min w^t C w subject to w^t c = 1

        Namely:
           w = u/c
        where u solves:
           min u^t C' u subject to u^t v = 1
        where
           v = [1,..,1]

        :param   C   Covariance
        :param   c   Linear coef as above
        :returns C'   Covariance playing role above

    """
    assert len(c) == np.shape(C)[0]
    assert all([ci > 0 for ci in c])
    cij = np.dot(np.atleast_2d(c), np.transpose(np.atleast_2d(c)))
    assert np.shape(cij) == np.shape(C)
    C_prime = C / cij
    return C_prime


def scaled_unitary_problem_solution(C, c):
    """
        Solve argmin   w^t C w  subject to w^t c = 1
    """
    Q, _ = scaled_unitary_problem(C=C, c=c)
    u = unitary_from_cov(cov=Q)
    w = u/ c
    assert np.dot(w, c) == 1
    return w


def left_multiplied_unitary_problem(C, R, c=None):
    """ Find equivalent problem to

          min w^t R C w   subject to w^t c ==1      (1)

        Instead find H^t H = R and set u = H w, w = H^{-1} u

          min  u^t  H  C  H^{-1} u    subject to u^t H^{-1}^t c = 1     (2)

        So set
                  c' = H^{-1}^t c
                  C' = H  C  H^{-1}

        Given  C', H, c' we can thus find
        w by solving the transformed problem (2), re-written:

           min u^t C' u subject to u^t c' = 1

        and then substituting back:

           w = H^{-1} u             (3)

        which is to say  w = solve(H,u)

        :param C   Covariance
        :param R   Additional tranform matrix


    """
    n = np.shape(C)[0]
    if c is None:
        c = np.ones(n)

    L = np.linalg.cholesky(R)
    H = np.transpose(L)
    assert np.linalg.norm( np.dot(np.transpose(L),L)-R ) < 1e-6
    c_prime = np.linalg.solve(L,c)
    C_Hinv = multiply_by_inverse(C,H)
    C_prime = np.dot( H, C_Hinv )

    return C_prime, H, c_prime


def left_multiplied_unitary_solution(C, R, c, verify=False):
    """ Find portfolio in a rather indirect way ... useful for checking

         min w^t R C w   subject to w^t c == 1
    """
    C_prime, H, c_prime = left_multiplied_unitary_problem(C=C, R=R)
    u = scaled_unitary_problem_solution(C=C_prime, c=c_prime)
    w = np.linalg.solve(H,u)

    if verify:
        RC = np.dot(R,C)
        w_check = scaled_unitary_problem_solution(C=RC,c=c)
        assert np.allclose( w, w_check )

    return w


def left_multiplied_unitary_reduction(C, R, c=None):
    """  Returns C'' and H and c'' such that if
             min u^t C'' u  subject to u^t 1 = 1
         and
             w = H^{-1} u/c'  = solve(H,u)/c''
         then w also solves
             min w^t R C w  subject to w^t c = 1

         (This unravels the transforms above)

         :return C', H, c'
    """
    C_intermediate, H, c_prime = left_multiplied_unitary_problem(C=C, R=R, c=c)
    C_prime = scaled_unitary_problem(C=C_intermediate, c=c_prime)
    return C_prime, H, c_prime


from precise.skaters.managers.weakmanagers import WEAK_LONG_MANAGERS
from precise.skaters.managers.ppomanagers import PPO_LONG_MANGERS
from precise.skaters.managers.hrpmanagers import HRP_LONG_MANAGERS
from precise.skaters.managers.schurmanagers import SCHUR_LONG_MANAGERS
from precise.skaters.managers.rflmanagers import RFL_HRP_LONG_MANAGERS
from tomark import Tomark
import random

# d0 managers unless otherwise stated

LONG_MANAGERS = WEAK_LONG_MANAGERS + PPO_LONG_MANGERS + HRP_LONG_MANAGERS + SCHUR_LONG_MANAGERS + RFL_HRP_LONG_MANAGERS
LS_MANAGERS = []
MANAGERS = LONG_MANAGERS + LS_MANAGERS


def manager_from_name(name):
    valid = [f for f in MANAGERS if f.__name__ == name]
    return valid[0] if len(valid)==1 else None


def manager_manifest():
    from precise.whereami import url_from_manager_name
    return dict([(f.__name__, url_from_manager_name(f.__name__)) for f in MANAGERS])


def random_cov_skater():
    return random.choice(MANAGERS)


def manager_manifest_markdown():
    manifest = manager_manifest()
    data = [ {'manager':mgr,'location':url} for mgr, url in manifest.items() ]
    markdown = Tomark.table(data)
    return markdown


def write_manager_manifest():
    from precise.whereami import MANAGER_MANIFEST
    markdown = manager_manifest_markdown()
    with open(MANAGER_MANIFEST,'wt') as fh:
        fh.write(markdown)



if __name__=='__main__':
    from pprint import pprint
    pprint(manager_manifest())
    write_manager_manifest()





from precise.skaters.portfolioutil.portcomparison import m6_equity_portfolio_correlation_rankings


def ports_test(ports, n_dim=10):
    """
          A quick crude way to test a slew of portfolio generation functions

          ports: List of portfolio generators
    """
    return m6_equity_portfolio_correlation_rankings(ports, n_dim=n_dim)



import numpy as np
import pandas as pd
from precise.skaters.covarianceutil.covfunctions import cov_to_corrcoef, nearest_pos_def
from precise.skaters.covarianceutil.datascatterfunctions import scatter_tensor_flat
from precise.skaters.covarianceutil.pdutil import data_to_square_dataframe


# Functions that take xs: (n_samples, n_vars) and produce cov matrices, or related
# Can also take a pd.DataFrame whose column names will then be used


def data_population_covariance(xs):
    """
        xs:  (n_samples, n_vars)
        Like np.cov but works if there is only one row
    """
    if isinstance(xs,pd.DataFrame):
        return data_to_square_dataframe(df=xs, data_func=data_population_covariance)
    else:
        if np.shape(xs)[0]==1:
            return np.var(np.array(xs).ravel())*np.eye(1)
        else:
            return np.cov(np.array(xs), rowvar=False, bias=True)


def data_population_correlation(xs):
    """
        Dimension of output is consistent.
    """
    if isinstance(xs, pd.DataFrame):
        return data_to_square_dataframe(df=xs, data_func=data_population_correlation)
    else:
        return cov_to_corrcoef(data_population_covariance(xs))


def scatter_func_cov(xs, cov_loc_func, demean=False, make_pos=True):
    """ Compute covariance using some location function acting on the flattenned scatter data
    :param xs:
    :param cov_loc_func:    Should take 2d array and return a column-wise pseudo-mean (i.e. axis=0)
    :param make_pos:        If True, will ensure cov matrix is pos def
    :param demean:
    :return:
    """
    if isinstance(xs, pd.DataFrame):
        return data_to_square_dataframe(df=xs, data_func=scatter_func_cov, cov_loc_func=cov_loc_func, demean=demean, make_pos=make_pos)
    else:
        n_samples, n_dim = np.shape(xs)
        if n_samples<=2:
            return np.eye(n_dim)
        else:
            stf = scatter_tensor_flat(xs=xs, demean=demean)  # (n_samples, n_dim*n_dim )
            assert np.shape(stf)[1]==n_dim*n_dim
            loc = cov_loc_func(stf)
            cov = np.reshape(loc, newshape=(n_dim, n_dim))
            if make_pos:
                cov = nearest_pos_def(a=cov)
            return cov


def scatter_skater_cov(xs, f, demean=False):
    """ Compute covariance using a univariate skater function on the flattened scatter data entries """

    if isinstance(xs, pd.DataFrame):
        return data_to_square_dataframe(df=xs, data_func=scatter_skater_cov, f=f, demean=demean)
    else:
        def skater_func(xs_):
            n_samples, n_entries = np.shape(xs_)
            final_x = list()
            for j in range(n_entries):
                s = {}
                x = np.squeeze(xs_[:, j])
                for y in x:
                    x, x_std, s = f(s=s, y=y, k=1)
                final_x.append(x)
            return np.array(final_x)

        return scatter_func_cov(xs=xs, cov_loc_func=skater_func, demean=demean)



from precise.skaters.covariance.ewaempfactory import ewa_emp_pcov_factory
from precise.skaters.covarianceutil.conventions import Y_DATA_TYPE
from precise.skaters.covarianceutil.differencing import d1_factory


def ewa_emp_pcov_d0_r01(y, s, k=1, e=1):
    return ewa_emp_pcov_factory(y=y, s=s, k=k, r=0.01)


def ewa_emp_pcov_d0_r02(y, s, k=1, e=1):
    return ewa_emp_pcov_factory(y=y, s=s, k=k, r=0.02)


def ewa_emp_pcov_d0_r05(y, s, k=1, e=1):
    return ewa_emp_pcov_factory(y=y, s=s, k=k, r=0.05)


def ewa_emp_pcov_d0_r10(y, s, k=1, e=1):
    return ewa_emp_pcov_factory(y=y, s=s, k=k, r=0.10)


EMA_DO_COV_SKATERS = [ewa_emp_pcov_d0_r01, ewa_emp_pcov_d0_r02, ewa_emp_pcov_d0_r05, ewa_emp_pcov_d0_r10]


def ewa_emp_pcov_d1_r01(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    """
        For when changes are iid gaussian
    """
    return d1_factory(f = ewa_emp_pcov_d0_r01, y=y, s=s, k=k)


def ewa_emp_pcov_d1_r02(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    return d1_factory(f = ewa_emp_pcov_d0_r02, y=y, s=s, k=k)


def ewa_emp_pcov_d1_r05(y:Y_DATA_TYPE, s:dict, k=1):
    return d1_factory(f = ewa_emp_pcov_d0_r05, y=y, s=s, k=k)


def ewa_emp_pcov_d1_r10(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    return d1_factory(f = ewa_emp_pcov_d0_r10, y=y, s=s, k=k)


EXP_EMP_D1_COV_SKATERS = [ewa_emp_pcov_d1_r01, ewa_emp_pcov_d1_r02, ewa_emp_pcov_d1_r05, ewa_emp_pcov_d1_r10]




import numpy as np
from precise.skaters.locationutil.vectorfunctions import scatter
from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory
from precise.skaters.portfolioutil.portfunctions import portfolio_variance


# Portfolio "lemmas" - not sure what else to call this just now ;)

def bAb(A,b):
    return np.dot( np.dot( np.transpose(b), A), b )


def quirky_solve(A, b):
    """
        Nuts alternative to np.linalg.solve(A,b)
    """
    v = np.ones(len(b))
    A_tilde = A / scatter(b)
    A_inv = np.linalg.inv(A)
    A_tilde_inv = np.linalg.inv( A_tilde )
    numer = bAb(A_inv, b)
    denom = bAb(A_tilde_inv, v)
    x = numer/denom*np.dot(A_tilde_inv, v) / b
    return x






from precise.skaters.portfoliostatic.ppoportfactory import ppo_portfolio_factory, PPO_LONG_BOUNDS, PPO_UNIT_BOUNDS

# Using PyPortfolioOpt, with some minor tweaks


def ppo_sharpe_long_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_sharpe', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_vol_long_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='min_volatility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_quad_long_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_quadratic_utility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_LONG_BOUNDS)


def ppo_sharpe_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_sharpe', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


def ppo_vol_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='min_volatility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


def ppo_quad_port(cov=None, pre=None, as_dense=True):
    return ppo_portfolio_factory(method='max_quadratic_utility', cov=cov, pre=pre, as_dense=as_dense, weight_bounds=PPO_UNIT_BOUNDS)


PPO_LONG_PORT = [ ppo_sharpe_long_port, ppo_vol_long_port, ppo_quad_long_port ]
PPO_LS_PORT = [ ppo_sharpe_port, ppo_vol_port, ppo_quad_port   ]
PPO_PORT =  PPO_LONG_PORT + PPO_LS_PORT


from precise.skaters.managers.schurmanagerfactory import schur_weak_pm_manager_factory, schur_weak_ewa_manager_factory


def hrp_weak_pm_t0_d0_r025_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, r=0.025, n_emp=50, e=e)


def hrp_weak_pm_t0_d0_r050_n25_long_manager(y, s, k=1, e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, r=0.050, n_emp=25, e=e)


def hrp_weak_emp_d0_r025_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, r=0.025, n_emp=50, e=e)


def hrp_weak_emp_d0_r050_n25_long_manager(y, s, k=1, e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, e=e, r=0.05, n_emp=25)




HRP_LONG_MANAGERS = [ hrp_weak_pm_t0_d0_r025_n50_long_manager,
                      hrp_weak_pm_t0_d0_r050_n25_long_manager,
                      hrp_weak_emp_d0_r025_n50_long_manager,
                      hrp_weak_emp_d0_r050_n25_long_manager
                      ]
HRP_LS_MANAGERS = []
HRP_MANAGERS = HRP_LONG_MANAGERS + HRP_LS_MANAGERS


from precise.skaters.covariance.ewapmfactory import ewa_pm_factory

# Partial moment skaters
# TODO: Make more autonomous
# Remark: Yes there are a few here and that's due to their initial success in Elo ratings.


def ewa_pm_emp_scov_r005_n100(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.005, n_emp=100)


def ewa_pm_emp_scov_r005_n100_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.005, n_emp=100, target=0)


def ewa_pm_emp_scov_r005_n200(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.005, n_emp=200)


def ewa_pm_emp_scov_r005_n200_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.005, n_emp=200, target=0)


def ewa_pm_emp_scov_r01_n100(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.01, n_emp=100)


def ewa_pm_emp_scov_r01_n100_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.01, n_emp=100, target=0)


def ewa_pm_emp_scov_r01_n200(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.01, n_emp=100)


def ewa_pm_emp_scov_r01_n200_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.01, n_emp=100, target=0)


def ewa_pm_emp_scov_r02_n50(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=50)


def ewa_pm_emp_scov_r02_n50_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=50, target=0)


def ewa_pm_emp_scov_r02_n100(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=100)


def ewa_pm_emp_scov_r02_n100_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=100, target=0)


def ewa_pm_emp_scov_r05_n50(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.05, n_emp=50)


def ewa_pm_emp_scov_r05_n50_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.05, n_emp=50, target=0)


def ewa_pm_emp_scov_r05_n25(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=25)


def ewa_pm_emp_scov_r05_n25_t0(s, y, k=1, e=1):
    return ewa_pm_factory(s=s, y=y, k=k, r=0.02, n_emp=25, target=0)



EWA_PM_EMP_D0_COV_SKATERS_TM = [ewa_pm_emp_scov_r01_n100, ewa_pm_emp_scov_r01_n200,
                             ewa_pm_emp_scov_r02_n100, ewa_pm_emp_scov_r01_n200,
                             ewa_pm_emp_scov_r02_n50, ewa_pm_emp_scov_r02_n100,
                             ewa_pm_emp_scov_r05_n50, ewa_pm_emp_scov_r05_n25 ]

EWA_PM_EMP_D0_COV_SKATERS_T0 = [ewa_pm_emp_scov_r01_n100_t0, ewa_pm_emp_scov_r01_n200_t0,
                             ewa_pm_emp_scov_r02_n100_t0, ewa_pm_emp_scov_r01_n200_t0,
                             ewa_pm_emp_scov_r02_n50_t0, ewa_pm_emp_scov_r02_n100_t0,
                             ewa_pm_emp_scov_r05_n50_t0, ewa_pm_emp_scov_r05_n25_t0 ]


EWA_PM_EMP_D0_COV_SKATERS = EWA_PM_EMP_D0_COV_SKATERS_TM + EWA_PM_EMP_D0_COV_SKATERS_T0




from precise.skaters.portfoliostatic.schurportfactory import hierarchical_seriation_portfolio_factory

# HRP treated as special case of Schur


def hierarchical_risk_parity_portfolio_factory(seriator=None, alloc=None, port=None, splitter=None, cov=None, pre=None,
                                               n_split=5):
    port_kwargs = {}
    return hierarchical_seriation_portfolio_factory(seriator=seriator, port=port, splitter=splitter, port_kwargs=port_kwargs,
                                                    cov=cov, pre=pre, n_split=n_split)





import numpy as np
from precise.skaters.covarianceutil.hubercov import scatter_huber_cov
from precise.skaters.covariance.buffactory import _buf
from precise.skaters.covarianceutil.conventions import X_TYPE


def buf_huber_d0_factory(s,y,n_buffer, a,b,n_iter=20, atol=1e-8, e=1):
    s = buf_mean_and_huber_pcov(s=s,x=y, a=a,b=b, n_iter=n_iter, atol=atol, n_buffer=n_buffer, e=e)
    try:
        x = s['mean']
        x_cov = s['pcov']
    except KeyError:
        x = np.array(y)
        x_cov = np.eye(len(y))

    return x, x_cov, s


def buf_mean_and_huber_pcov(s:dict=None, x:X_TYPE=None, n_buffer:int=100,a=1.0, b=2.0,n_iter=20, atol=1e-8, e=1)->dict:
    # Similar to np.cov(xs[max(0, k - n_buffer + 1):k + 1], axis=0) but uses Huber location for cov estimate

    def hcov(xs_):
        return scatter_huber_cov(xs=xs_, a=a,b=b, n_iter=n_iter, atol=atol, demean=True)

    return _buf(funcs=[np.nanmean, hcov], func_names=['mean', 'pcov'], func_kwargs=[{'axis':0}, {}], s=s, x=x, n_buffer=n_buffer, e=e)



import numpy as np
from precise.skaters.covarianceutil.conventions import X_TYPE, X_DATA_TYPE, is_data


# Vectorized median for finite buffer
# Equivalent to np.nanmedian(xs[max(0,k-n_buffer+1):k+1], axis=0)


def med(s:dict=None, x:X_TYPE=None, n_buffer:int=100 ):
    if not s:
        s = _med_init(s=s, n_buffer=n_buffer)
    if is_data(x):
        s = _med_update(s=s, x=x)
    return s


def _med_init(s:dict=None, n_buffer:int=None):
    if not s:
        s = dict()
    s.update( {'buffer':list(), 'n_buffer':n_buffer} )
    return s


def _med_update(s:dict, x:X_DATA_TYPE, n_buffer:int=None):
    if n_buffer is None:
        n_buffer = s['n_buffer']
    if len(s['buffer']):
        assert len(s['buffer'][0])==len(x), 'dimension mismatch'
    s['buffer'].append(np.atleast_1d(x))
    if len(s['buffer'])>n_buffer:
        _ = s['buffer'].pop(0)
    s['median'] = np.nanmedian(s['buffer'],axis=0)
    return s



if __name__=='__main__':
    import random
    xs = np.random.randn(500,100)
    xs[17,1] = np.nan
    s = {}
    n_buffer = random.choice([1,5,100])
    for k,x in enumerate(xs):
        s = med(s=s,x=x,n_buffer=n_buffer)
        np_median = np.nanmedian(xs[max(0,k-n_buffer+1):k+1], axis=0)
        s_median = s['median']
        if not np.allclose( np_median, s_median, atol=1e-8 ):
            print('urgh')
            assert False



from precise.skaters.covarianceutil.covfunctions import try_invert, weaken_cov
from precise.skaters.portfoliostatic.unitportfactory import unitary_from_cov
from precise.skaters.portfolioutil.portfunctions import exclude_negative_weights, portfolio_variance
import scipy
from numpy.linalg import LinAlgError

# Fast long-only approximately min-var portfolios where the only constraints are sum(w)=1, w>0


def weak_portfolio_factory(cov=None, pre=None, a=1.0, b=None, with_neg_mass=False):
    """ Long only portfolio somewhat similar to unitary min-var portfolio
        Uses a "weakenned" cov matrix

             sum(w)=1
             w>=0

    :param cov: Covariance matrix
    :param pre: Precision matrix (either cov or pre must be supplied)
    :param a:   Weakenning parameter
    :param b:   Weakenning parameter
    :param with_neg_mass:
    :return:
    """
    if cov is None:
        cov = try_invert(pre)
    w0 = unitary_from_cov(cov=cov)
    if b is not None:
        return _weak_known_params(cov=cov, a=a, b=b, w0=w0, with_neg_mass=with_neg_mass)
    else:
        return _weak_optimal_b(cov=cov, w0=w0, a=a, with_neg_mass=with_neg_mass)


def _weak_known_params(cov, a, b, w0, with_neg_mass=False):
    w1 = _weak_from_cov(cov=cov, a=a, b=b, w=w0, with_weak=False)
    return exclude_negative_weights(w1, with_neg_mass=with_neg_mass)


def _weak_optimal_b(cov, w0, a, with_neg_mass=False):

    def b_objective(u,w,a, v0):
        try:
            w1 = _weak_from_cov(cov, a=a, b=u[0], w=w, with_weak=False)
            v = portfolio_variance(cov=cov, w=exclude_negative_weights(w1))/v0
            return v
        except LinAlgError:
            bad_v = portfolio_variance(cov=cov, w=100*w)
            return bad_v

    v0 = portfolio_variance(cov=cov,w=w0)
    res = scipy.optimize.minimize(fun=b_objective,x0=0.75, bounds=[(0,1)], args=(w0, a, v0))
    best_b = res.x[0]
    best_w = _weak_from_cov(cov=cov, a=1.0, b=best_b, w=w0, with_weak=False)
    return exclude_negative_weights(w=best_w,with_neg_mass=with_neg_mass)


def _weak_from_cov(cov, w, a=1.0, b=0.75, with_weak=False):
    """
        Given long-short weights w, find different portfolio using weaker cov
        This is likely to have a lot less negative mass

        :param a  Diagnonal multipier for neg weights
        :param b  Additional multiplier for off-diag
    """
    d = [a if wi < 0.0 else 1.0 for wi in w]
    weak_cov = weaken_cov(cov=cov, diag_multipliers=d, off_diag_additional_factor=b)
    w1 = unitary_from_cov(weak_cov)
    return (w1, weak_cov) if with_weak else w1






from precise.skaters.portfoliostatic.diagportfactory import diagonal_portfolio_factory
from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory
from precise.skaters.portfoliostatic.schurportfactory import schur_portfolio_factory
from precise.skaters.portfoliostatic.diagalloc import diag_alloc
from precise.skaters.portfoliostatic.unitalloc import unit_alloc
from precise.skaters.portfoliostatic.weakalloc import weak_long_alloc

# Schur complement portfolios 

# Some that use the same portfolio constructor as allocation...


def schur_unit_unit_s5_g100_port(cov=None, pre=None):
    return schur_portfolio_factory(port=unit_portfolio_factory, alloc=unit_alloc, cov=cov, pre=pre, n_split=5, gamma=1.0)


def schur_weak_weak_s5_g100_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=weak_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=1.0)


def schur_diag_diag_s5_g100_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=diagonal_portfolio_factory, alloc=diag_alloc, cov=cov, pre=pre, n_split=5, gamma=1.0)


def schur_unit_unit_s5_g050_port(cov=None, pre=None):
    return schur_portfolio_factory(port=unit_portfolio_factory, alloc=unit_alloc, cov=cov, pre=pre, n_split=5, gamma=0.5)


def schur_weak_weak_s5_g050_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=weak_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=0.5)


def schur_diag_diag_s5_g050_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=diagonal_portfolio_factory, alloc=diag_alloc, cov=cov, pre=pre, n_split=5, gamma=0.5)


# Using weak allocation..


def schur_unit_weak_s5_g100_port(cov=None, pre=None):
    return schur_portfolio_factory(port=unit_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=1.0)


def schur_diag_weak_s5_g100_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=diagonal_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=1.0)


def schur_unit_weak_s5_g050_port(cov=None, pre=None):
    return schur_portfolio_factory(port=unit_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=0.5)


def schur_diag_weak_s5_g050_long_port(cov=None, pre=None):
    return schur_portfolio_factory(port=diagonal_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5, gamma=0.5)



SCHUR_LONG_PORT = [schur_diag_diag_s5_g100_long_port, schur_weak_weak_s5_g100_long_port,
                   schur_diag_diag_s5_g050_long_port, schur_weak_weak_s5_g050_long_port]
SCHUR_LS_PORT = [schur_unit_unit_s5_g100_port, schur_unit_unit_s5_g050_port]
SCHUR_PORT = SCHUR_LONG_PORT + SCHUR_LS_PORT



if __name__=='__main__':
    from precise.skaters.portfolioutil.portcomparison import m6_equity_portfolio_variance_rankings
    from precise.skaters.portfoliostatic.hrpport import HRP_PORT
    from pprint import pprint
    rankings = m6_equity_portfolio_variance_rankings(ports=SCHUR_PORT + HRP_PORT, n_dim=50)
    pprint(rankings)


from precise.skaters.covariance.runemp import RUN_EMP_DO_COV_SKATERS, RUN_EMP_D1_COV_SKATERS
from precise.skaters.covariance.bufemp import BUF_EMP_D0_SKATERS, BUF_EMP_D1_SKATERS
from precise.skaters.covariance.bufsk import BUF_SK_D0_SKATERS, BUF_SK_D1_SKATERS
from precise.skaters.covariance.ewaemp import EMA_DO_COV_SKATERS, EXP_EMP_D1_COV_SKATERS
from precise.skaters.covariance.bufhuber import BUF_HUBER_D0_COV_SKATERS, BUF_HUBER_D1_COV_SKATERS
from precise.skaters.covariance.ewapm import EWA_PM_EMP_D0_COV_SKATERS
from precise.skaters.covariance.ewalw import EWA_LW_D0_COV_SKATERS, EWA_LW_D1_COV_SKATERS
from precise.skaters.covariance.ewalz import EWA_LZ_D0_COV_SKATERS
import random
from precise.whereami import COV_SKATER_MANIFEST
from tomark import Tomark

# List of fully autonomous multivariate gaussian forecasters
# Run this file to print a list of skaters and their code URLs

ALL_D0_SKATERS = BUF_EMP_D0_SKATERS + \
                 RUN_EMP_DO_COV_SKATERS + \
                 BUF_SK_D0_SKATERS + \
                 EMA_DO_COV_SKATERS + \
                 BUF_HUBER_D0_COV_SKATERS + \
                 EWA_PM_EMP_D0_COV_SKATERS + \
                 EWA_LW_D0_COV_SKATERS+\
                 EWA_LZ_D0_COV_SKATERS


ALL_D1_SKATERS = BUF_EMP_D1_SKATERS + \
                 RUN_EMP_D1_COV_SKATERS + \
                 BUF_SK_D1_SKATERS + \
                 EXP_EMP_D1_COV_SKATERS + \
                 BUF_HUBER_D1_COV_SKATERS + \
                 EWA_LW_D1_COV_SKATERS

ALL_COV_SKATERS = ALL_D0_SKATERS + ALL_D1_SKATERS


def cov_skater_from_name(name):
    valid = [f for f in ALL_COV_SKATERS if f.__name__ == name]
    return valid[0] if len(valid)==1 else None


def cov_skater_manifest():
    from precise.whereami import url_from_skater_name
    return dict([(f.__name__, url_from_skater_name(f.__name__)) for f in ALL_COV_SKATERS])


def random_cov_skater():
    return random.choice(ALL_D0_SKATERS)


def cov_skater_manifest_markdown():
    manifest = cov_skater_manifest()
    data = [ {'covariance skater':mgr,'location':url} for mgr, url in manifest.items() ]
    markdown = Tomark.table(data)
    return markdown


def write_cov_skater_manifest():
    markdown = cov_skater_manifest_markdown()
    with open(COV_SKATER_MANIFEST,'wt') as fh:
        fh.write(markdown)


if __name__=='__main__':
    from pprint import pprint
    pprint(cov_skater_manifest())
    write_cov_skater_manifest()



# Utilities for when data is presented as (n_samples, n_vars)

import numpy as np


def scatter_tensor(xs, demean=False):
    """ Returns all the individual scatter matrices

    :param xs:           (n_samples, n_vars)
    :param demean:
    :return: (n_samples, n_vars, n_vars )
    """
    if demean:
        b = xs - np.mean(np.array(xs), axis=0)
    else:
        b = xs
    b_rows = np.atleast_3d(b)
    b_cols = np.transpose(b_rows, axes=[0, 2, 1])
    sc = np.matmul(b_rows, b_cols)
    return sc


def scatter_tensor_flat(xs, demean=False):
    """ Flat scatter
    :param xs:  (n_samples, n_vars)
    :param demean:
    :return:    (n_samples, n_vars*n_vars )
    """
    sc = scatter_tensor(xs=xs, demean=demean)
    [n_samples, n_dim, n_dim_check] = np.shape(sc)
    assert n_dim == n_dim_check
    return np.reshape(sc, newshape=[n_samples, n_dim * n_dim])



if __name__=='__main__':
    a = np.random.randn(10,3)
    scatter_tensor(a, demean=True)


from precise.skaters.covariance.buffactory import buf_mean_and_pcov
from precise.skaters.covarianceutil.conventions import Y_DATA_TYPE
import numpy as np

def buf_emp_pcov_d0_factory(y:Y_DATA_TYPE, s:dict, n_buffer:int=100, e=1):
    """
        Empirical covariance skater that assumes y's are iid

    :param y:  (n_dim,)   Incoming vector of observations
    :param s             State
    :param k:             Steps ahead to predict (this won't change anything)
    :param a:             Variables know in advance (ignored)
    :param t:             Epoch time (ignored)
    :param e:             Allocated computation time (ignored)
    :return:  (n_dim,), (n_dim,n_dim), dict
              Point estimate, cov estimate, posterior state
    """
    s = buf_mean_and_pcov(s=s, x=y, n_buffer=n_buffer, e=e)
    try:
        x = s['mean']
    except KeyError:
        x = y

    try:
        x_cov = s['pcov']
    except KeyError:
        x_cov = np.eye(len(y))

    return x, x_cov, s



from precise.skaters.managers.managerfactory import static_cov_manager_factory_d0
from precise.skaters.covariance.ewapm import ewa_pm_factory, ewa_pm_emp_scov_r01_n100
from precise.skaters.portfoliostatic.ppoportfactory import ppo_portfolio_factory, PPO_LONG_BOUNDS
from precise.skaters.covariance.ewaempfactory import ewa_emp_pcov_factory
from functools import partial


def ppo_long_manager_factory(y,s, f, method, e=1):
    port = partial(ppo_portfolio_factory, method=method, as_dense=True, weight_bounds=PPO_LONG_BOUNDS)
    return static_cov_manager_factory_d0(f=f, port=port, y=y, s=s, e=e)


def ppo_pm_long_manager_factory(y, s, method, target, n_emp, r, e=1):
    """
       PyPortfolioOpt portfolio construction using partial moments cov estimation
    """
    f = partial( ewa_pm_factory, k=1,r=r,target=target, n_emp=n_emp )
    port = partial( ppo_portfolio_factory, method=method, as_dense=True, weight_bounds=PPO_LONG_BOUNDS )
    return static_cov_manager_factory_d0(f=f, port=port, y=y, s=s, e=e)


def ppo_ewa_long_manager_factory(y, s, method, n_emp, r, e=1):
    """
       PyPortfolioOpt portfolio construction using EWA cov estimation
    """
    f = partial( ewa_emp_pcov_factory, k=1, r=r, n_emp=n_emp )
    port = partial( ppo_portfolio_factory, method=method, as_dense=True, weight_bounds=PPO_LONG_BOUNDS )
    return static_cov_manager_factory_d0(f=f, port=port, y=y, s=s, e=e)




from precise.skaters.portfoliostatic.ppoport import PPO_LS_PORT, PPO_LONG_PORT
from precise.skaters.portfoliostatic.weakport import WEAK_LONG_PORT, WEAK_LS_PORT
from precise.skaters.portfoliostatic.diagport import DIAG_LONG_PORT
from precise.skaters.portfoliostatic.hrpport import HRP_LONG_PORT, HRP_LS_PORT
from precise.skaters.portfoliostatic.schurport import SCHUR_LONG_PORT, SCHUR_LS_PORT
from precise.skaters.portfoliostatic.unitport import UNIT_LS_PORT, UNIT_LONG_PORT
from precise.skaters.portfoliostatic.equalport import EQUAL_LONG_PORT, EQUAL_LS_PORT

import random

LONG_PORT = PPO_LONG_PORT + WEAK_LONG_PORT + DIAG_LONG_PORT + HRP_LONG_PORT + UNIT_LONG_PORT + SCHUR_LONG_PORT + EQUAL_LONG_PORT
LS_PORT = PPO_LS_PORT + HRP_LS_PORT + WEAK_LS_PORT + SCHUR_LS_PORT + UNIT_LS_PORT + EQUAL_LS_PORT
PORT = LONG_PORT + LS_PORT


def random_port():
    return random.choice(PORT)


if __name__=='__main__':
    from precise.skaters.portfolioutil.portcomparison import stock_portfolio_variance_points_race
    report = stock_portfolio_variance_points_race(n_iter=5000, ports=LONG_PORT, n_dim=500, n_obs=100, k=5, n_top=100)
    from pprint import pprint
    pprint(report)


import pandas_datareader.data as web
import pandas as pd
import time
from functools import lru_cache
import numpy as np
import time

@lru_cache(maxsize=500)
def get_prices(ticker,n_obs,interval, max_attempts=10):
    print('Getting '+ticker)
    time.sleep(5)
    success = False
    attempts = 0
    while not success:
        try:
            data = web.get_data_yahoo(ticker, interval=interval)[-n_obs - 1:]['Close'].values
            success = True
        except Exception as e:
            print(str(e))
            print('backing off')
            time.sleep(10)
            attempts += 1
            if attempts>=max_attempts:
                raise ValueError
    return data



def live_equity_returns(tickers, n_obs=60, interval='m', k=1):
    df = pd.DataFrame()
    for ticker in tickers:
        try:
            data = get_prices(ticker=ticker, n_obs=n_obs, interval=interval)
            assert len(data)==n_obs+1
            values = list(np.diff(np.log(data),k))[::k]
            df[ticker] = values
        except:
            pass
    return df


def live_veteran_etf_data(interval='d',k=1):
    from precise.skatertools.data.etflists import VETERAN_NON_BOND_ETFS
    n_obs = 60 if interval=='m' else 250
    df = live_equity_returns(tickers=VETERAN_NON_BOND_ETFS, n_obs=n_obs, interval=interval, k=k)
    return df


def random_m6_returns(n_dim=10, n_obs:int=60, verbose=True, interval='m', etf=0, **ignore):
    """
        Use portfolio for M6 competition
    """
    constituents = pd.read_csv('https://raw.githubusercontent.com/microprediction/m6/main/data/official/M6_Universe.csv')
    if etf:
        tickers = constituents['symbol']
    else:
        tickers = constituents['symbol'][50:]
    return random_equity_returns(all_tickers=tickers, n_dim=n_dim, n_obs=n_obs, verbose=verbose, interval=interval)


def random_equity_returns(all_tickers, n_dim=10, n_obs:int=60, verbose=True, interval='m', **ignore):
    """ Get return series
    :param n_dim:            Number of assets
    :param n_obs:
    :param interval: 'd' or 'm'
    :return: prices NOT necessarily corresponding to all_tickers
    """
    assert 2*n_dim<=len(all_tickers)
    if interval=='m':
        assert n_obs<=60,'too many observations for monthly'

    tickers = list(np.random.choice(all_tickers, n_dim * 2, replace=False))
    prices = list()
    while (len(prices)<n_dim) and len(tickers):
        ticker = tickers.pop()
        data = []
        try:
            data = get_prices(ticker=ticker,n_obs=n_obs, interval=interval)
            if len(data)==n_obs+1:
                prices.append(np.diff(np.log(data)))
            if verbose:
                print('Got '+ticker+' len '+str(len(data)))
        except Exception as e:
            print('Failure getting '+ticker)
            time.sleep(1)

    prices_transposed = list(map(list, zip(*prices)))
    return np.array(prices_transposed)



if __name__=='__main__':
    import numpy as np
    from pprint import pprint
    a = random_m6_returns(n_dim=3, n_obs=60, etf=1)
    c = np.corrcoef(np.array(a),rowvar=False)
    pprint(c)


from precise.skaters.covariance.ewalzfactory import ewa_emp_lz_pcov_factory_rg_rl_n
from precise.skaters.covarianceutil.differencing import d1_factory

# EWALZ short for "Exp Weight Avg" Lee-Zhong


def ewa_lz_scov_d0_rg01_rl01_n100(y, s, k=1, e=1):
    assert k==1
    return ewa_emp_lz_pcov_factory_rg_rl_n(y=y, s=s, rg=0.01, rl=0.01, n=100, e=e)


def ewa_lz_scov_d0_rg005_rl01_n100(y, s, k=1, e=1):
    assert k==1
    return ewa_emp_lz_pcov_factory_rg_rl_n(y=y, s=s, rg=0.005, rl=0.01, n=10, e=e)


def ewa_lz_scov_d0_rg005_rl02_n50(y, s, k=1, e=1):
    assert k==1
    return ewa_emp_lz_pcov_factory_rg_rl_n(y=y, s=s, rg=0.005, rl=0.02, n=50, e=e)




EWA_LZ_D0_COV_SKATERS = [ ewa_lz_scov_d0_rg01_rl01_n100,
                          ewa_lz_scov_d0_rg005_rl01_n100,
                          ewa_lz_scov_d0_rg005_rl02_n50]




import time
import random
import pandas as pd

# Replica of https://raw.githubusercontent.com/microprediction/timemachines/main/timemachines/skatertools/data/skaterresiduals.py to avoid timemachine dependency

SKATER_RESIDUAL_URL = 'https://raw.githubusercontent.com/microprediction/precisedata/main/skaterresiduals/skater_residuals_0.csv'
n_data = 450


def random_skater_residual_dataframe(n_obs:int):
    got = False
    while not got:
        the_choice = random.choice(list(range(n_data)))
        the_url = SKATER_RESIDUAL_URL.replace('N', str(the_choice))
        try:
            df = pd.read_csv(the_url)
            del df['Unnamed: 0']
            got = len(df.index) > n_obs + 10
        except:
            got = False
    return df

def random_long_residual(n_obs):
    """ Returns random long time series of skater residuals """
    assert n_obs<35000

    start_time = time.time()
    t = [ start_time+24*60*60*60*i for i in range(n_obs) ]
    df = random_skater_residual_dataframe(n_obs)
    col = random.choice(list(df.columns))
    vals = list(df[col].values)
    k = random.choice(list(range(0,len(df.index)-n_obs-5)))
    y = vals[k:k+n_obs]
    return y,t


def random_multivariate_residual(n_obs, as_dataframe=True, random_start=True):
    """ Canned skater residual data, potentially useful for studying ensembles """
    assert n_obs < 35000
    df = random_skater_residual_dataframe(n_obs=n_obs)
    if random_start:
        k = random.choice(list(range(0, len(df.index) - n_obs - 5)))
        df = df[k:k+n_obs]
    df.dropna(axis=0, how='any', inplace=True)
    return df[:n_obs] if as_dataframe else df[:n_obs].values


def random_noncollinear_residual(n_obs, random_start=True):
    from collinearity import SelectNonCollinear
    xs = random_multivariate_residual(n_obs=n_obs, as_dataframe=False, random_start=random_start)
    selector = SelectNonCollinear(correlation_threshold=0.99)
    xs_noncollinear = selector.fit_transform(xs)
    return xs_noncollinear




if __name__=='__main__':
    y,t = random_long_residual(n_obs=1000)
    print(y[:10])


from precise.skaters.portfoliostatic.diagportfactory import diagonal_portfolio_factory
from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory
from precise.skaters.portfoliostatic.hrpportfactory import hierarchical_risk_parity_portfolio_factory
from precise.skaters.portfoliostatic.diagalloc import diag_alloc
from precise.skaters.portfoliostatic.unitalloc import unit_alloc
from precise.skaters.portfoliostatic.weakalloc import weak_long_alloc


# Hierarchical Risk-Parity


def hrp_diag_diag_s5_long_port(cov=None, pre=None):
    # Lopez de Prado 2016
    # Allocate using diagonal
    return hierarchical_risk_parity_portfolio_factory(port=diagonal_portfolio_factory, alloc=diag_alloc, cov=cov, pre=pre, n_split=5)


def hrp_unit_unit_s5_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=unit_portfolio_factory, alloc=unit_alloc, cov=cov, pre=pre, n_split=5)


def hrp_weak_weak_s5_long_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=weak_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5)


# Weak allocation ...

def hrp_unit_weak_s5_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=unit_portfolio_factory, alloc=unit_alloc, cov=cov, pre=pre, n_split=5)


def hrp_diag_weak_s5_long_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=weak_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5)


# Diag allocation, but different portfolios

def hrp_unit_diag_s5_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=unit_portfolio_factory, alloc=unit_alloc, cov=cov, pre=pre, n_split=5)


def hrp_weak_diag_s5_long_port(cov=None, pre=None):
    return hierarchical_risk_parity_portfolio_factory(port=weak_portfolio_factory, alloc=weak_long_alloc, cov=cov, pre=pre, n_split=5)



HRP_LONG_PORT = [hrp_weak_weak_s5_long_port, hrp_diag_diag_s5_long_port, hrp_diag_weak_s5_long_port, hrp_weak_diag_s5_long_port]
HRP_LS_PORT = [hrp_unit_unit_s5_port, hrp_unit_weak_s5_port, hrp_unit_diag_s5_port]
HRP_PORT = HRP_LONG_PORT + HRP_LS_PORT



if __name__=='__main__':
    from precise.skaters.portfolioutil.portcomparison import m6_equity_portfolio_correlation_rankings
    from pprint import pprint
    rankings = m6_equity_portfolio_correlation_rankings(ports=HRP_PORT, n_dim=40)
    pprint(rankings)


import numpy as np
from precise.skaters.covarianceutil.covfunctions import multiply_diag
from kmeans1d import cluster


def centroid_precision_adjacency(pre:np.ndarray)->np.ndarray:
    """ Uses clustering of absolute values to infer precision adjacency matrix

    :param  pre: 2d array of precision estimates
    :return: 2d array of boolean

    """
    # A rather speculative approach, this using 1-dim version of k-means clustering to establish a
    # minimum absolute precision value. Any entry above this threshold is considered 'real'.
    a = np.copy(pre)
    n = np.shape(pre)[0]
    off_diag_values = sorted(np.abs(multiply_diag(a, phi=0, copy=True).ravel()))
    clusters, centroids = cluster(array=off_diag_values, k=3)
    assert clusters[1]>=clusters[0]
    cutoff_value = min( centroids[1]*2.0, centroids[2]/2.0 )
    abs_pre = np.abs(pre)
    for i in range(n):
        a[:,i] = np.vectorize(int)(abs_pre[:,i] > cutoff_value)
    for i in range(n):
        a[i,i]=1
    return a.astype(bool)








from typing import Union
import numpy as np

X_TYPE = Union[list,np.ndarray,int]
X_DATA_TYPE = Union[list,np.ndarray]
Y_TYPE = X_TYPE
Y_DATA_TYPE = X_DATA_TYPE


def infer_dimension(n_dim:int=None, x:X_TYPE=None, **ignore)->int:
    """ Infer the number of variables
    :param n_dim:
    :param x:
    :return:
    """
    if n_dim is not None:
        return n_dim
    elif isinstance(x, int):
        return x
    elif len(x)>1:
        return len(x)
    else:
        raise ValueError('Ambiguity in number of variables. Try supplying x or n_dim')


def is_data(x):
    return isinstance(x,np.ndarray) or isinstance(x,list)


# Miscellaneous lists OF ETFS for paper

SOFI_ETFS = ['SFY', 'SFYX']
VANGUARD_ETFS = ['BND', 'VCIT', 'VOO', 'VTI', 'VUG', 'VV', 'VTV', 'VIG', 'VO', 'VB']  # <-- Those with fee <= 5bps
SCHWAB_ETFS = ['SCHB', 'SCHK', 'SCHX', 'SCHG', 'SCHV', 'SCHM', 'SCHA', 'SCHF', 'SCHC', 'SCHY', 'SCHE', 'SCHP', 'SCHR',
               'SCHJ', 'SCHI', 'SCHQ']
INVESCO_ETFS = ['SOXQ', 'IBBQ', 'RPV']
FACTOR_ETFS = ['PDP', 'SDOG', 'SPYG', 'SPLV', 'DGRW']
MISC_ETFS = ['RSP', 'NOBL', 'PGX', 'BKNL', 'FNDE', 'CWB', 'RPG', 'GNR', 'SGOL', 'XLG', 'CWI', 'HYS', 'GXC',
             'VOTE', 'GBUG']

ETFS = list(set(SOFI_ETFS + VANGUARD_ETFS + SCHWAB_ETFS + INVESCO_ETFS + FACTOR_ETFS + MISC_ETFS))

# A few with decent history via yahoo datareader:
VETERAN_NON_BOND_ETFS = ['RPG', 'SCHC', 'SCHV', 'VUG', 'SDOG', 'VOO', 'DGRW', 'CWB', 'CWI', 'SCHA', 'VTV', 'VIG',
                'NOBL', 'SCHG', 'SCHX', 'SPLV', 'VO', 'VTI', 'FNDE', 'SCHM', 'VV', 'XLG', 'GXC',
                'PDP', 'GNR', 'SPYG', 'RSP', 'RPV',  'SCHE', 'SCHF', 'SCHB', 'VB']
VETERAN_BOND_ETFS = ['HYS', 'BND','SCHP', 'SCHR','PGX','VCIT']
VETERAN_ETFS = VETERAN_NON_BOND_ETFS + VETERAN_BOND_ETFS

VETERAN_VOL = {'BND': 6,
               'CWB': 24,
               'CWI': 17,
               'DGRW': 22,
               'FNDE': 17,
               'GNR': 17,
               'GXC': 23.5,
               'HYS': 9,
               'NOBL': 16,
               'PDP': 27,
               'PGX': 17,
               'RPG': 34,
               'RPV': 29,
               'RSP': 23,
               'SCHA': 27,
               'SCHB': 17,
               'SCHC': 17,
               'SCHE': 21,
               'SCHF': 19,
               'SCHG': 32,
               'SCHM': 25,
               'SCHP': 9.3,
               'SCHR': 4.8,
               'SCHV': 21,
               'SCHX': 22,
               'SDOG': 17,
               'SGOL': 15,
               'SPLV': 20,
               'SPYG': 30,
               'VB': 26,
               'VCIT': 7,
               'VIG': 20,
               'VO': 25,
               'VOO': 25,
               'VTI': 25,
               'VTV': 20,
               'VUG': 30,
               'VV': 26,
               'XLG': 27}


def veterans():
    from precise.skatertools.data.equitylive import get_prices
    import time
    veteran_etfs = list()
    for etf in ETFS:
        try:
            data = get_prices(ticker=etf, interval='m', n_obs=60, max_attempts=1)
            n = len(data)
            print((etf, str(n)))
            if n == 61:
                veteran_etfs.append(etf)
            time.sleep(5)
        except:
            print('No data for ' + etf)
            pass
    return veteran_etfs


if __name__ == '__main__':
    print(len(VETERAN_ETFS))
    print([missing for missing in VETERAN_ETFS if not missing in VETERAN_VOL])



import numpy as np
from precise.skaters.covarianceutil.conventions import X_TYPE, X_DATA_TYPE, is_data, infer_dimension

# Running means for vectors


def ewa(s:dict=None, x:X_TYPE=None, r:float=0.025, n_dim:int=None):
    """ Exponential weighted average """
    return averager(s=s, x=x, r=r, n_dim=n_dim, method='ma')


def sma(s:dict=None, x:X_TYPE=None, r:float=0.025, n_dim:int=None):
    """ Switching from empirical to ewa """
    return averager(s=s, x=x, r=r, n_dim=n_dim, method='switch')



def averager(s:dict=None, x:X_TYPE=None, r:float=0.025, method:str= 'switch', n_dim:int=None):
    """ Stands for "switching moving average"
          method:   'switch', 'ma' or 'emp'
    """
    if not s:
        s = _sma_init(s=s,x=x,r=r,n_dim=n_dim, method=method)
    if is_data(x):
        s = _sma_update(s=s,x=x,r=r, method=method)
    return s


def _sma_init(s:dict=None, x:X_TYPE=None, n_dim:int=None, r=None, method:str=None):
    n_dim = infer_dimension(n_dim=n_dim, x=x)
    if not s:
        s = dict()
    s.update( {'n_samples':0,'mean':np.zeros(n_dim),'r':r,'method':method})
    return s


def _sma_update(s:dict, x:X_DATA_TYPE, r=None, method:str=None):
    method = method or s['method']
    r = r if (r is not None) else s['r']
    s['n_samples'] += 1
    if (s['n_samples']<1/r and method=='switch') or (method=='emp'):
        s['mean'] += x/s['n_samples']
    else:
        if s['n_samples']==1:
            s['mean'] = x
        else:
            s['mean'] = (1-r)*s['mean'] + r*x
    return s






from precise.skaters.locationutil.hubermean import huber_mean
from precise.skaters.covarianceutil.datafunctions import scatter_func_cov
import numpy as np
from pprint import pprint
import pandas as pd

# Covariance estimation using a generalized Huber pseudo-mean
# Here xs can be a pd.DataFrame


def scatter_huber_cov(xs, a:float=1.0, b=2.0, n_iter=20, atol=1e-8, demean=False):
    """ Huber cov """
    def hloc(xs_):
        return huber_mean(xs=xs_, a=a, b=b, n_iter=n_iter, atol=atol)

    return scatter_func_cov(xs=xs, cov_loc_func=hloc, demean=demean)


def scatter_emp_cov(xs, demean=False):
    """ Empirical cov """
    # Here mostly as a check

    def traditional_mean(xs):
        return xs[0] if len(xs )==1 else np.nanmean(xs ,axis=0)

    return scatter_func_cov(xs=xs, cov_loc_func=traditional_mean, demean=demean)


if __name__=='__main__':
    xs = np.random.randn(20,4)
    ce = scatter_emp_cov(xs=xs, demean=True)
    ch = scatter_huber_cov(xs=xs, demean=True)
    pprint(ce)
    pprint(ch)
    df = pd.DataFrame(data=xs)
    dfc = scatter_huber_cov(xs=df)
    print(dfc)




# Functions that respect typical key conventions
#
# i.e.   f(**s) usually works where s is the state for a covariance skater

import numpy as np


def oas(n_samples:int, pcov=None, scov=None, **ignore):
    """ Compute shrunk covariance matrix from empirical
    :param pcov:         Covariance matrix
    :return:             Shrunk cov matrix
    """
    # In sklearn this is bundled with emp cov estimation, so we have to cut and paste a few lines. See
    # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09b/sklearn/covariance/_shrunk_covariance.py#L347
    assert n_samples > 1, 'Need n_samples>1'
    if pcov is None:
        assert scov is not None, 'Need pcov or scov to be supplied'
        pcov =  (n_samples-1)/n_samples*scov

    n_dim = np.shape(pcov)[0]
    mu = np.trace(pcov) / n_dim
    alpha = np.mean(pcov ** 2)
    num = alpha + mu ** 2
    den = (n_samples + 1.0) * (alpha - (mu ** 2) / n_dim)
    shrinkage = 1.0 if den == 0 else min(num / den, 1.0)
    b = (1.0 - shrinkage) * pcov
    b.flat[:: n_dim + 1] += shrinkage * mu
    return b



from typing import List
from precise.skaters.portfoliostatic.weakallocfactory import weak_allocation_factory

# List of allocators that use cov weakening


def weak_long_alloc(covs:List, pres:List=None)->[float]:
    """ Allocate capital between portfolios using either cov or pre matrices
    :param covs:  List of covariance matrices
    :param pres:  List of precision matrices
    :return: Capital allocation vector
    """
    return weak_allocation_factory(covs=covs, pres=pres)


WEAK_ALLOC = [weak_long_alloc]


import numpy as np
from precise.skaters.covarianceutil.conventions import infer_dimension, X_DATA_TYPE, is_data


def emp_pcov(s:dict, x:[float]=None, n_dim=None, k=1)->dict:
    """
        Track empirical sample covariance
    """
    assert k==1
    if not s:
        s = _emp_pcov_init(x=x,n_dim=n_dim)
    if is_data(x):
        s = _emp_pcov_update(s=s, x=x)
    return s


def _emp_pcov_init(s:dict=None, x:X_DATA_TYPE=None, n_dim=None):
    """ Empirical population covariance"""
    n_dim = infer_dimension(x=x,n_dim=n_dim)
    if s is None:
       s = dict()
    s['n_dim']=n_dim
    s['shape']=(n_dim, n_dim)
    s['ones'] = np.ones(n_dim)
    s['n_samples'] = 0
    s['mean'] = np.zeros(n_dim)
    s['pcov'] = np.eye(n_dim)
    return s


def _emp_pcov_update(s:dict, x:X_DATA_TYPE, target=None):
    assert s['n_dim'] == len(x)
    prev_cov = np.copy( s['pcov'] )
    prev_mean = s['mean']
    s['n_samples'] += 1
    s['mean'] = prev_mean + (x - prev_mean)/s['n_samples']
    if target is None:
        delta_x_prev = np.atleast_2d(x-prev_mean)
        delta_x_current = np.atleast_2d(x - s['mean'])
    else:
        delta_x_prev = np.atleast_2d(x - target)
        delta_x_current = np.atleast_2d(x - target)
    s['pcov'] = prev_cov + ( np.matmul( delta_x_current.transpose(),delta_x_prev) - prev_cov ) / s['n_samples']

    return s


def merge_emp_scov(s:dict, other_s:dict):
    """ Merge two online covariance tracking objects as if the data had been merged """
    if other_s['n_dim'] != s['n_dim']:
        raise ValueError(
            f'''
               Cannot merge two OnlineCovariances with different orders.
               ({s['n_dim']} != {other_s['n_dim']})
               ''')

    merged_cov = _emp_pcov_init(n_dim=s['n_dim'])
    merged_cov['n_samples'] = s['n_samples'] + other_s['n_samples']
    count_corr = (other_s['n_samples'] * s['n_samples']) / merged_cov['n_samples']
    merged_cov['mean'] = (s['mean'] / other_s['n_samples'] + other_s['mean'] / s['n_samples']) * count_corr
    flat_mean_diff = s['mean'] - other_s['mean']
    mean_diffs = np.broadcast_to(flat_mean_diff, s['shape']).T
    merged_cov['pcov'] = (s['pcov'] * s['n_samples']  + other_s['pcov'] * other_s['n_samples'] + mean_diffs * mean_diffs.T * count_corr) / merged_cov['n_samples']
    return merged_cov